"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber import io, bedio, scriptio, seqio
from fastx_barber.const import FastxFormats
from fastx_barber.exception import enable_rich_assert
from fastx_barber.match import search_needle
from fastx_barber.scripts import arguments as ap
import logging
import os
from rich.logging import RichHandler  # type: ignore

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        __name__.split(".")[-1],
        description="Scan a FASTX file for a sequence.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Scan a FASTX file a sequence and generate a BED file with its locations.",
    )

    parser.add_argument(
        "input",
        type=str,
        metavar="in.fastx[.gz]",
        help="""Path to the fasta/q file to scan for matches.""",
    )
    parser.add_argument(
        "needle",
        type=str,
        help="""Sequence to scan for.""",
    )

    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced.add_argument(
        "--output",
        type=str,
        metavar="out.bed[.gz]",
        help="Path to fasta/q file where to write trimmed records. "
        + "Format will match the input. Defaults to input file with BED extension.",
    )
    advanced.add_argument(
        "--prefix",
        type=str,
        metavar="prefix",
        help="""Name prefix. Default: 'loc_'""",
        default="loc_",
    )
    advanced.add_argument(
        "--case-insensitive",
        action="store_const",
        dest="case_insensitive",
        const=True,
        default=False,
        help="Perform case-insensitive scan.",
    )
    advanced.add_argument(
        "--global-name",
        action="store_const",
        dest="global_name",
        const=True,
        default=False,
        help="Global location name. Requires sorted FASTA.",
    )
    advanced = ap.add_compress_level_option(advanced)
    advanced = ap.add_log_file_option(advanced)

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    assert os.path.isfile(args.input), f"file not found: '{args.input}'"

    fmt, _ = seqio.get_fastx_format(args.input)
    assert fmt in [
        FastxFormats.FASTA,
        FastxFormats.FASTQ,
    ], f"input must be a FASTX file. ({fmt})"

    if args.log_file is not None:
        scriptio.add_log_file_handler(args.log_file)

    if args.output is None:
        base, ext, gzipped = io.is_gzipped(args.input)
        args.output = f"{base}.bed"
        if gzipped:
            args.output += ".gz"

    if args.case_insensitive:
        args.needle = args.needle.upper()

    return args


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    logging.info("[bold underline red]General[/]")
    logging.info(f"Input\t\t{args.input}")
    logging.info(f"Needle\t\t{args.needle}")
    logging.info(f"Output\t\t{args.output}")
    logging.info(f"Prefix\t\t{args.prefix}")
    logging.info(f"Global\t\t{args.global_name}")

    parser, fmt = seqio.get_fastx_parser(args.input)

    writer = bedio.BedWriter(args.output, 4)

    location_id = 0
    for record in parser:
        if args.case_insensitive:
            record[1] = record[1].upper()
        for pos, location_id in search_needle(record, args.needle, location_id):
            writer.do(
                (
                    record[0],
                    pos,
                    pos + len(args.needle) + 1,
                    f"{args.prefix}{location_id}",
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                    None,
                )
            )
        if not args.global_name:
            location_id = 0

    logging.info("Done. :thumbs_up: :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from abc import ABCMeta, abstractmethod
from Bio import SeqIO  # type: ignore
from fastx_barber.const import FlagData
from fastx_barber.io import is_gzipped
from fastx_barber.const import FastxFormats, FastxExtensions
import gzip
import os
from typing import Any, Dict, IO, Iterator, List, Optional, Set, Tuple, Type, Union

SimpleFastxRecord = Tuple[str, str, Optional[str]]
SimpleFastaRecord = Tuple[str, str, None]
SimpleFastqRecord = Tuple[str, str, str]

SimpleFastxParser = Union[
    SeqIO.QualityIO.FastqGeneralIterator, SeqIO.FastaIO.SimpleFastaParser
]


def get_fastx_format(path: str) -> Tuple[FastxFormats, bool]:
    """
    Returns:
        Tuple[FastxFormats, bool] -- fastx file format and gzipped status
    """
    base, ext, gzipped = is_gzipped(path)
    assert FastxExtensions.has_value(ext), f"Unrecognized extension '{ext}'."
    if ext in FastxExtensions.FASTA.value:
        return (FastxFormats.FASTA, gzipped)
    elif ext in FastxExtensions.FASTQ.value:
        return (FastxFormats.FASTQ, gzipped)
    else:
        return (FastxFormats.NONE, False)


def get_fastx_parser(path: str) -> Tuple[SimpleFastxParser, FastxFormats]:
    """Retrieves appropriate simple parser and associated format (fasta or fastq)."""
    fmt, gzipped = get_fastx_format(path)
    handle: Union[str, IO] = path
    if gzipped:
        handle = gzip.open(path, "rt")
    else:
        handle = open(path, "r+")
    assert fmt in FastxFormats
    if FastxFormats.FASTA == fmt:
        parser = SeqIO.FastaIO.SimpleFastaParser(handle)
        parser = ((*x, None) for x in parser)
    elif FastxFormats.FASTQ == fmt:
        parser = SeqIO.QualityIO.FastqGeneralIterator(handle)
    else:
        parser = (x for x in handle)
    return (parser, fmt)


class FastxChunkedParser(object):
    """Parser with chunking capabilities for fasta and fastq files.

    Variables:
        __IH: SimpleFastxParser {[type]} -- [description]
    """

    __IH: SimpleFastxParser
    __chunk_size: int
    __chunk_counter: int = 0

    def __init__(self, parser: SimpleFastxParser, chunk_size: int):
        super(FastxChunkedParser, self).__init__()
        self.__IH = parser
        assert chunk_size > 0
        self.__chunk_size = chunk_size

    @property
    def chunk_size(self):
        return self.__chunk_size

    @property
    def last_chunk_id(self):
        return self.__chunk_counter

    def __next__(self) -> Tuple[List[SimpleFastxRecord], int]:
        chunk: List[SimpleFastxRecord] = []
        while len(chunk) < self.__chunk_size:
            try:
                chunk.append(next(self.__IH))
            except StopIteration:
                break
        if 0 == len(chunk):
            raise StopIteration
        else:
            self.__chunk_counter += 1
            return (chunk, self.__chunk_counter)

    def __iter__(self) -> Iterator[Tuple[List[SimpleFastxRecord], int]]:
        return self


class ABCSimpleWriter(metaclass=ABCMeta):
    """Simple record writer abstract base class

    Extends:
        metaclass=ABCMeta

    Variables:
        _OH {IO} -- output buffer handle
    """

    _OH: IO

    def __init__(self, path: str, compress_level: int = 6):
        """Initialize simple writer

        Arguments:
            path {str} -- path to output file

        Keyword Arguments:
            compress_level {int} -- gzip compression level (default: {6})
        """
        super(ABCSimpleWriter, self).__init__()
        _, _, gzipped = is_gzipped(path)
        if gzipped:
            self._OH = gzip.open(path, "wt", compress_level)
        else:
            self._OH = open(path, "w+")

    @property
    def name(self) -> str:
        return self._OH.name

    @abstractmethod
    def write(self, record: Any, *args) -> None:
        """Write record to output buffer

        Decorators:
            abstractmethod

        Arguments:
            record {Any} -- record to be written
        """
        pass

    def close(self):
        """Close output buffer handle"""
        self._OH.close()

    @staticmethod
    def close_handle(self):
        if isinstance(self, ABCSimpleWriter):
            self.close()


class SimpleFastxWriter(ABCSimpleWriter):

    _fmt: FastxFormats

    def __init__(self, path: str, compress_level: int = 6):
        super(SimpleFastxWriter, self).__init__(path, compress_level)
        self._fmt, _ = get_fastx_format(path)
        assert self._fmt in FastxFormats

    @property
    def format(self) -> FastxFormats:
        return self._fmt

    @abstractmethod
    def write(self, record: SimpleFastxRecord, *args) -> None:
        """Write record to output buffer

        Decorators:
            abstractmethod

        Arguments:
            record {SimpleFastxRecord} -- record to be written
        """
        pass


class SimpleFastaWriter(SimpleFastxWriter):
    def __init__(self, path: str, compress_level: int = 6):
        super(SimpleFastaWriter, self).__init__(path, compress_level)
        assert FastxFormats.FASTA == self.format

    def write(self, record: SimpleFastxRecord, *args) -> None:
        self._OH.write(f">{record[0]}\n{record[1]}\n")


class SimpleFastqWriter(SimpleFastxWriter):
    def __init__(self, path: str, compress_level: int = 6):
        super(SimpleFastqWriter, self).__init__(path, compress_level)
        assert FastxFormats.FASTQ == self.format

    def write(self, record: SimpleFastxRecord, *args) -> None:
        self._OH.write(f"@{record[0]}\n{record[1]}\n+\n{record[2]}\n")


def get_fastx_writer(fmt: FastxFormats) -> Type[SimpleFastxWriter]:
    """Retrieves appropriate simple writer class."""
    if FastxFormats.FASTA == fmt:
        return SimpleFastaWriter
    elif FastxFormats.FASTQ == fmt:
        return SimpleFastqWriter
    return SimpleFastxWriter


class ABCSimpleSplitWriter(metaclass=ABCMeta):

    _base_path: str
    _root_path: str
    _basename: str
    _split_key: str
    _split_by: Set[str]
    _compress_level: int
    _is_gzipped: bool

    def __init__(self, path: str, split_key: str, compress_level: int = 6):
        super(ABCSimpleSplitWriter, self).__init__()
        self._base_path = path
        self._root_path = os.path.dirname(path)
        self._basename = os.path.basename(path)
        if self._basename.startswith("."):
            self._basename = self._basename[1:]
        self._is_gzipped = self._basename.endswith(".gz")
        self._split_key = split_key
        self._split_by = set()
        self._compress_level = compress_level

    @property
    def split_key(self):
        return self._split_key

    @property
    def split_by(self):
        return self._split_by

    def opened_before(self, split_value: str) -> bool:
        return split_value in self._split_by

    def open(self, split_value: str) -> IO:
        path = os.path.join(
            self._root_path,
            f"{self._split_key}_split.{split_value}.{self._basename}",
        )
        if self._is_gzipped:
            if self.opened_before(split_value):
                return gzip.open(path, "at", self._compress_level)
            else:
                self._split_by.add(split_value)
                return gzip.open(path, "wt", self._compress_level)
        else:
            if self.opened_before(split_value):
                return open(path, "a")
            else:
                self._split_by.add(split_value)
                return open(path, "w")

    @abstractmethod
    def write(self, record: Any, flag_data: Dict[str, FlagData], *args) -> None:
        """Write record to output buffer

        Decorators:
            abstractmethod

        Arguments:
            record {Any} -- record to be written
            flag_data {Dict[str, FlagData]} -- flag data for splitting
        """
        pass

    def close(self):
        pass


class SimpleSplitFastxWriter(ABCSimpleSplitWriter):

    _fmt: FastxFormats

    def __init__(self, path: str, split_key: str, compress_level: int = 6):
        super(SimpleSplitFastxWriter, self).__init__(path, split_key, compress_level)
        self._fmt, _ = get_fastx_format(path)
        assert self._fmt in FastxFormats

    @property
    def format(self) -> FastxFormats:
        return self._fmt

    @abstractmethod
    def write(
        self, record: SimpleFastxRecord, flag_data: Dict[str, FlagData], *args
    ) -> None:
        """Write record to output buffer

        Decorators:
            abstractmethod

        Arguments:
            record {SimpleFastxRecord} -- record to be written
            flag_data {Dict[str, FlagData]} -- flag data for splitting
        """
        pass


class SimpleSplitFastaWriter(SimpleSplitFastxWriter):
    def __init__(self, path: str, split_key: str, compress_level: int = 6):
        super(SimpleSplitFastaWriter, self).__init__(path, split_key, compress_level)
        assert FastxFormats.FASTA == self.format

    def write(
        self, record: SimpleFastxRecord, flag_data: Dict[str, FlagData], *args
    ) -> None:
        assert (
            self._split_key in flag_data
        ), f"Cannot split by flag '{self._split_key}'. Flag not found."
        OH = self.open(flag_data[self._split_key][0])
        OH.write(f">{record[0]}\n{record[1]}\n")
        OH.close()


class SimpleSplitFastqWriter(SimpleSplitFastxWriter):
    def __init__(self, path: str, split_key: str, compress_level: int = 6):
        super(SimpleSplitFastqWriter, self).__init__(path, split_key, compress_level)
        assert FastxFormats.FASTQ == self.format

    def write(
        self, record: SimpleFastxRecord, flag_data: Dict[str, FlagData], *args
    ) -> None:
        assert (
            self._split_key in flag_data
        ), f"Cannot split by flag '{self._split_key}'. Flag not found."
        OH = self.open(flag_data[self._split_key][0])
        OH.write(f"@{record[0]}\n{record[1]}\n+\n{record[2]}\n")
        OH.close()


def get_split_fastx_writer(fmt: FastxFormats) -> Type[SimpleSplitFastxWriter]:
    """Retrieves appropriate simple writer class."""
    if FastxFormats.FASTA == fmt:
        return SimpleSplitFastaWriter
    elif FastxFormats.FASTQ == fmt:
        return SimpleSplitFastqWriter
    return SimpleSplitFastxWriter



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from enum import Enum
from typing import DefaultDict, Optional, Tuple

__version__ = "0.1.3"


BedRecord = Tuple[
    str,
    int,
    int,
    Optional[str],
    Optional[float],
    Optional[str],
    Optional[int],
    Optional[int],
    Optional[str],
    Optional[int],
    Optional[int],
    Optional[int],
]
BedExtension = ".bed"


class FastxFormats(Enum):
    """Fastx formats

    Extends:
        Enum

    Variables:
        FASTA {str}
        FASTQ {str}
        NONE {str} -- Not fastx
    """

    FASTA = "fasta"
    FASTQ = "fastq"
    NONE = "None"

    @classmethod
    def has_value(self, value):
        return value in self._value2member_map_


class FastxExtensions(Enum):
    """Fastx extensions

    Extends:
        Enum

    Variables:
        FASTA {tuple}
        FASTQ {tuple}
    """

    FASTA = (".fa", ".fasta")
    FASTQ = (".fq", ".fastq")

    @classmethod
    def has_value(self, value):
        return any(value in v for v in self._value2member_map_)


QFLAG_START = "q"
DEFAULT_PHRED_OFFSET = 33

PATTERN_EXAMPLE = "^(?<UMI>.{8})(?<BC>.{8})(?<CS>GATC){s<2}"

"""Flag data, contains matched str, start, and end position"""
FlagData = Tuple[str, int, int]
FlagStatsType = DefaultDict[str, DefaultDict[str, int]]

# Unit tests related stuff
UT_FLAG_NAME = "fake"
UT_RECORD_SEQ_LEN = 200
UT_N_RECORDS = 100
UT_CHUNK_SIZE = 8
UT_FLAG_PATTERN = f"^(?<{UT_FLAG_NAME}>.{{8}})"



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber import io
import os
import shutil
import tempfile


def test_check_tmp_dir():
    path = io.check_tmp_dir()
    assert os.path.isdir(path)
    shutil.rmtree(path)


def test_is_gzipped():
    assert io.is_gzipped("test.txt.gz")[2]
    assert not io.is_gzipped("test.txt")[2]


def test_splitext():
    base, ext = io.splitext("/root/test.txt")
    assert "/root/test" == base
    assert ".txt" == ext
    base, ext = io.splitext("/root/test.txt.gz")
    assert "/root/test" == base
    assert ".txt.gz" == ext
    base, ext = io.splitext("/root/test.txt.gz.gz")
    assert "/root/test" == base
    assert ".txt.gz.gz" == ext


def test_ChunkMerger():
    TD = tempfile.TemporaryDirectory()

    C1 = open(os.path.join(TD.name, ".tmp.chunk1.test.txt"), "w+")
    C1.write("chunk1\n")
    C1.close()

    C2 = open(os.path.join(TD.name, ".tmp.chunk2.test.txt"), "w+")
    C2.write("chunk2\n")
    C2.close()

    merger = io.ChunkMerger(TD)
    merger.do("test.txt", 2, "test_description")

    assert not os.path.isfile(C1.name)

    MH = open("test.txt", "r+")
    merged_content = MH.readlines()
    assert 2 == len(merged_content)
    assert "chunk1\n" == merged_content[0]
    assert "chunk2\n" == merged_content[1]
    MH.close()
    os.remove(MH.name)

    shutil.rmtree(TD.name)


def test_ChunkMerger_split():
    TD = tempfile.TemporaryDirectory()

    C = open(os.path.join(TD.name, "test_split.ASD.tmp.chunk1.test.txt"), "w+")
    C.write("chunk1\n")
    C.close()

    C = open(os.path.join(TD.name, "test_split.DSA.tmp.chunk1.test.txt"), "w+")
    C.write("chunk2\n")
    C.close()

    C = open(os.path.join(TD.name, "test_split.ASD.tmp.chunk2.test.txt"), "w+")
    C.write("chunk3\n")
    C.close()

    C = open(os.path.join(TD.name, "test_split.DSA.tmp.chunk2.test.txt"), "w+")
    C.write("chunk4\n")
    C.close()

    merger = io.ChunkMerger(TD, "test")
    merger.do("test.txt", 2, "test_description")

    assert not os.path.isfile(C.name)

    MH = open("test_split.ASD.test.txt", "r+")
    merged_content = MH.readlines()
    print(merged_content)
    assert 2 == len(merged_content)
    assert "chunk1\n" == merged_content[0]
    assert "chunk3\n" == merged_content[1]
    MH.close()
    os.remove(MH.name)

    MH = open("test_split.DSA.test.txt", "r+")
    merged_content = MH.readlines()
    print(merged_content)
    assert 2 == len(merged_content)
    assert "chunk2\n" == merged_content[0]
    assert "chunk4\n" == merged_content[1]
    MH.close()
    os.remove(MH.name)

    shutil.rmtree(TD.name)


def test_ChunkMerger_noRemove():
    TD = tempfile.TemporaryDirectory()

    C1 = open(os.path.join(TD.name, ".tmp.chunk1.test.txt"), "w+")
    C1.write("chunk1\n")
    C1.close()

    C2 = open(os.path.join(TD.name, ".tmp.chunk2.test.txt"), "w+")
    C2.write("chunk2\n")
    C2.close()

    merger = io.ChunkMerger(TD)
    merger.do_remove = False
    merger.do("test.txt", 2, "test_description")

    assert os.path.isfile(C1.name)

    MH = open("test.txt", "r+")
    merged_content = MH.readlines()
    assert 2 == len(merged_content)
    assert "chunk1\n" == merged_content[0]
    assert "chunk2\n" == merged_content[1]
    MH.close()
    os.remove(MH.name)

    shutil.rmtree(TD.name)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber import scripts
from fastx_barber.scripts import arguments as ap
import sys


def default_parser(*args) -> None:
    print("fbarber flag -h for usage details.")
    sys.exit()


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        __name__.split(".")[-1],
        description="""FASTX flag tools.""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="FASTX flag tools.",
    )
    parser.set_defaults(parse=default_parser)
    parser = ap.add_version_option(parser)

    sub_subparsers = parser.add_subparsers(
        title="sub-commands",
        help="Access the help page for a sub-command with: sub-command -h",
    )

    scripts.flag_extract.init_parser(sub_subparsers)
    scripts.flag_filter.init_parser(sub_subparsers)
    scripts.flag_regex.init_parser(sub_subparsers)
    scripts.flag_split.init_parser(sub_subparsers)
    scripts.flag_stats.init_parser(sub_subparsers)

    return parser



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber.const import __version__, DEFAULT_PHRED_OFFSET
import joblib  # type: ignore
import logging
import sys
import tempfile


def log_args(args: argparse.Namespace) -> None:
    logging.info("[bold underline red]General[/]")
    logging.info(f"Input\t\t{args.input}")
    logging.info(f"Pattern\t\t{args.pattern.pattern}")
    logging.info(f"Threads\t\t{args.threads}")
    logging.info(f"Chunk size\t{args.chunk_size}")


def add_version_option(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser.add_argument(
        "--version", action="version", version=f"{sys.argv[0]} {__version__}"
    )
    return parser


def add_unmatched_output_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--unmatched-output",
        type=str,
        help="""Path to fasta/q file where to write records that do not match
        the pattern. Format must match the input.""",
    )
    return arg_group


def add_compress_level_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--compress-level",
        type=int,
        default=6,
        help="""GZip compression level. Default: 6.""",
    )
    return arg_group


def add_log_file_option(arg_group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--log-file", type=str, help="""Path to file where to write the log."""
    )
    return arg_group


def add_chunk_size_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--chunk-size",
        type=int,
        default=50000,
        help="""How many records per chunk. Default: 50000""",
    )
    return arg_group


def add_threads_option(arg_group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--threads",
        type=int,
        default=1,
        help="""Threads for parallelization. Default: 1""",
    )
    return arg_group


def add_tempdir_option(arg_group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--temp-dir",
        type=str,
        help="""Path to temporary folder.""",
        default=tempfile.gettempdir(),
    )
    return arg_group


def check_threads(threads: int) -> int:
    if threads > joblib.cpu_count():
        return joblib.cpu_count()
    elif threads <= 0:
        return 1
    return threads


def add_phred_offset_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--phred-offset",
        type=str,
        default=DEFAULT_PHRED_OFFSET,
        help="""Phred offset for qscore calculation.""",
    )
    return arg_group


def add_flag_delim_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--flag-delim",
        type=str,
        default="~",
        help="""Delimiter for flags. Used twice for flag separation and once
        for key-value pairs. It should be a single character. Default: '~'.
        Example: header~~flag1key~flag1value~~flag2key~flag2value""",
    )
    return arg_group


def add_filter_qual_flags_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--filter-qual-flags",
        type=str,
        nargs="+",
        help="""Space-separated 'flag_name,min_qscore,max_perc' strings, where
        bases with qscore < min_qscore are considered low quality, and max_perc
        is the largest allowed fraction of low quality bases. I.e., you can specify
        multiple flag filters by separating them with a space.""",
    )
    return arg_group


def add_filter_qual_output_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--filter-qual-output",
        type=str,
        help="""Path to fasta/q file where to write records that do not pass the
        flag filters. Format must match the input.""",
    )
    return arg_group


def add_comment_space_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--comment-space",
        type=str,
        default=" ",
        help="""Delimiter for header comments. Defaults to a space.""",
    )
    return arg_group


def add_split_by_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--split-by",
        type=str,
        help="""Flag to be used to split records in separate output files,
        based on flag value.""",
    )
    return arg_group


def add_flagstats_option(
    arg_group: argparse._ArgumentGroup,
) -> argparse._ArgumentGroup:
    arg_group.add_argument(
        "--flagstats",
        type=str,
        nargs="+",
        help="""Space-separate names of flags to calculate statistics for. By default
        this is skipped. Statistics are calculated before any quality filter, on records
        matching the provided pattern.""",
    )
    return arg_group



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber import match, qual, trim
from fastx_barber.const import FastxFormats
import regex as re  # type: ignore


def test_FastaTrimmer_trim_len():
    trimmer = trim.FastaTrimmer()
    record = ("test", "ATCGATCGATCGATCGATGCAT", None)

    trimmed_record = trimmer.trim_len(record, 5, 5)
    assert record[0] == trimmed_record[0]
    assert record[1][5:] == trimmed_record[1]
    assert record[2] == trimmed_record[2]

    trimmed_record = trimmer.trim_len(record, 5, 3)
    assert record[0] == trimmed_record[0]
    assert record[1][:-5] == trimmed_record[1]
    assert record[2] == trimmed_record[2]

    trimmed_record = trimmer.trim_len(record, 100, 3)
    assert record[0] == trimmed_record[0]
    assert "" == trimmed_record[1]
    assert record[2] == trimmed_record[2]


def test_FastaTrimmer_trim_re():
    trimmer = trim.FastaTrimmer()
    record = ("test", "ATCGATCGATCGATCGATGCAT", None)

    matcher = match.FastxMatcher(re.compile("^.*CGATCGA"))
    trimmed_record = trimmer.trim_re(record, matcher.do(record)[0])
    assert record[0] == trimmed_record[0]
    assert "TGCAT" == trimmed_record[1]
    assert record[2] == trimmed_record[2]

    matcher = match.FastxMatcher(re.compile("^.*?CGATCGA"))
    trimmed_record = trimmer.trim_re(record, matcher.do(record)[0])
    assert record[0] == trimmed_record[0]
    assert "TCGATCGATGCAT" == trimmed_record[1]
    assert record[2] == trimmed_record[2]


def test_FastqTrimmer_trim_len():
    trimmer = trim.FastqTrimmer()
    record = ("test", "ATCGATCGATCGATCGATGCAT", "A/A/A/A/A/A/A/A/A/A/A/")

    trimmed_record = trimmer.trim_len(record, 5, 5)
    assert record[0] == trimmed_record[0]
    assert record[1][5:] == trimmed_record[1]
    assert record[2][5:] == trimmed_record[2]

    trimmed_record = trimmer.trim_len(record, 5, 3)
    assert record[0] == trimmed_record[0]
    assert record[1][:-5] == trimmed_record[1]
    assert record[2][:-5] == trimmed_record[2]

    trimmed_record = trimmer.trim_len(record, 100, 3)
    assert record[0] == trimmed_record[0]
    assert "" == trimmed_record[1]
    assert "" == trimmed_record[2]


def test_FastqTrimmer_trim_re():
    trimmer = trim.FastqTrimmer()
    record = ("test", "ATCGATCGATCGATCGATGCAT", "A/A/A/A/A/A/A/A/A/A/A/")

    matcher = match.FastxMatcher(re.compile("^.*CGATCGA"))
    trimmed_record = trimmer.trim_re(record, matcher.do(record)[0])
    assert record[0] == trimmed_record[0]
    assert "TGCAT" == trimmed_record[1]
    assert "/A/A/" == trimmed_record[2]

    matcher = match.FastxMatcher(re.compile("^.*?CGATCGA"))
    trimmed_record = trimmer.trim_re(record, matcher.do(record)[0])
    assert record[0] == trimmed_record[0]
    assert "TCGATCGATGCAT" == trimmed_record[1]
    assert "/A/A/A/A/A/A/" == trimmed_record[2]


def test_FastqTrimmer_trim_qual():
    trimmer = trim.FastqTrimmer()
    qio = qual.QualityIO()
    record = ("test", "ATCGATCGATCGATCGATGCAT", "////A/A/A/A/A/A/A/////")

    trimmed_record, trimmed_len = trimmer.trim_qual(record, 30, 5, qio)
    assert 4 == trimmed_len
    assert record[0] == trimmed_record[0]
    assert "ATCGATCGATCGATGCAT" == trimmed_record[1]
    assert "A/A/A/A/A/A/A/////" == trimmed_record[2]

    trimmed_record, trimmed_len = trimmer.trim_qual(record, 30, 3, qio)
    assert 5 == trimmed_len
    assert record[0] == trimmed_record[0]
    assert "ATCGATCGATCGATCGA" == trimmed_record[1]
    assert "////A/A/A/A/A/A/A" == trimmed_record[2]


def test_get_fastx_trimmer():
    assert isinstance(
        trim.get_fastx_trimmer(FastxFormats.FASTA), type(trim.FastaTrimmer)
    )
    assert isinstance(
        trim.get_fastx_trimmer(FastxFormats.FASTQ), type(trim.FastqTrimmer)
    )
    assert isinstance(trim.get_fastx_trimmer(FastxFormats.NONE), type(trim.ABCTrimmer))



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from abc import ABCMeta, abstractmethod
from collections import defaultdict
from fastx_barber.const import FastxFormats, FlagData, FlagStatsType, QFLAG_START
from fastx_barber.match import ANPMatch
from fastx_barber.seqio import SimpleFastxRecord
import logging
import os
import pandas as pd  # type: ignore
import regex as re  # type: ignore
from rich.progress import track  # type: ignore
from typing import Any, Dict, List, Match, Optional, Pattern, Tuple, Type, Union


class FlagStats(object):

    __stats: FlagStatsType
    _flags_for_stats: Optional[List[str]] = None

    def __init__(self, flags_for_stats: Optional[List[str]] = None):
        super(FlagStats, self).__init__()
        self.__stats = defaultdict(lambda: defaultdict(lambda: 0))
        self._flags_for_stats = flags_for_stats

    def update(self, flags: Dict[str, FlagData]) -> None:
        if self._flags_for_stats is None:
            return
        for flag_name, flag_data in flags.items():
            if flag_name in self._flags_for_stats:
                self.__stats[flag_name][flag_data[0]] += 1

    def __getitem__(self, key):
        return self.__stats[key]

    def __setitem__(self, key, value):
        self.__stats[key] = value

    def keys(self):
        return self.__stats.keys()

    def values(self):
        return self.__stats.values()

    def items(self):
        return self.__stats.items()

    def get_dataframe(self, flag_name: str) -> pd.DataFrame:
        stats = self.__stats[flag_name]
        df = pd.DataFrame()
        df["value"] = list(stats.keys())
        df["counts"] = list(stats.values())
        df["perc"] = round(df["counts"] / df["counts"].sum() * 100, 2)
        df.sort_values("counts", ascending=False, ignore_index=True, inplace=True)
        return df

    def export(self, output_path: str, verbose: bool = True) -> None:
        output_dir = os.path.dirname(output_path)
        basename = os.path.basename(output_path)
        if basename.endswith(".gz"):
            basename = basename.split(".gz")[0]
        basename = os.path.splitext(basename)[0]
        if verbose:
            flag_keys = track(self.keys(), description="Exporting flagstats")
        else:
            flag_keys = self.keys()

        for flag_name in list(flag_keys):
            self.get_dataframe(flag_name).to_csv(
                os.path.join(output_dir, f"{basename}.{flag_name}.stats.tsv"),
                sep="\t",
                index=False,
            )


class ABCFlagBase(metaclass=ABCMeta):
    """Class with basic flag-related variables

    Extends:
        metaclass=ABCMeta

    Variables:
        _flag_delim {str} -- flag delimiter
        _comment_space {str} -- fastx comment separator
    """

    _flag_delim: str = "~"
    _comment_space: str = " "

    def __init__(self):
        super(ABCFlagBase, self).__init__()

    @property
    def flag_delim(self):
        return self._flag_delim

    @flag_delim.setter
    def flag_delim(self, flag_delim: str):
        assert 1 == len(flag_delim)
        self._flag_delim = flag_delim

    @property
    def comment_space(self):
        return self._flag_delim

    @comment_space.setter
    def comment_space(self, comment_space: str):
        assert 1 == len(comment_space)
        self._comment_space = comment_space


class ABCFlagExtractor(ABCFlagBase):
    """Flag extractor abstract base class

    Extends:
        ABCFlagBase

    Variables:
        _selected_flags {Optional[List[str]]} -- flags to extract
        _flags_for_stats {Optional[List[str]]} -- list of flags for stats calculation
        _flagstats {FlagStats} -- to contain flagstats generated by update_stats
    """

    _selected_flags: Optional[List[str]] = None
    _flagstats: FlagStats

    def __init__(
        self,
        selected_flags: Optional[List[str]] = None,
        flags_for_stats: Optional[List[str]] = None,
    ):
        self._selected_flags = selected_flags
        self._flagstats = FlagStats(flags_for_stats)

    @property
    def flagstats(self):
        return self._flagstats

    @abstractmethod
    def extract_selected(
        self, record: Any, match: Union[ANPMatch, Match, None]
    ) -> Dict[str, FlagData]:
        """Extract selected flags

        Flags are selected according to self._selected_flags

        Decorators:
            abstractmethod

        Arguments:
            record {Any} -- record from where to extract flags
            match {Match} -- results of matching the record to a flag pattern

        Returns:
            Dict[str, FlagData] -- a dictionary with flag name as key and data as value
        """
        pass

    @abstractmethod
    def extract_all(
        self, record: Any, match: Union[ANPMatch, Match, None]
    ) -> Dict[str, FlagData]:
        """Extract all flags

        Decorators:
            abstractmethod

        Arguments:
            record {Any} -- record from where to extract flags
            match {Match} -- results of matching the record to a flag pattern

        Returns:
            Dict[str, FlagData] -- a dictionary with flag name as key and data as value
        """
        pass

    @abstractmethod
    def update(self, record: Any, flag_data: Dict[str, FlagData]) -> Any:
        """Update record

        Decorators:
            abstractmethod

        Arguments:
            record {Any} -- record to update based on flags
            flag_data {Dict[str, FlagData]} -- a dictionary with flag name as key
                                               and data as value

        Returns:
            Any -- updated record.
        """
        pass

    def update_stats(self, flags: Dict[str, FlagData]) -> None:
        self._flagstats.update(flags)

    def apply_selection(self, flag_data: Dict[str, FlagData]) -> Dict[str, FlagData]:
        """Subselects provided flags.

        According to self._selected_flags

        Decorators:
            abstractmethod

        Arguments:
            flag_data {Dict[str, FlagData]} -- a dictionary with flag name as key
                                               and data as value
        """
        if self._selected_flags is None:
            return flag_data
        else:
            selected_flag_data = {}
            for name in self._selected_flags:
                if name in flag_data.keys():
                    selected_flag_data[name] = flag_data[name]
            return selected_flag_data


class FastaFlagExtractor(ABCFlagExtractor):
    def __init__(
        self,
        selected_flags: Optional[List[str]] = None,
        flags_for_stats: Optional[List[str]] = None,
    ):
        super(FastaFlagExtractor, self).__init__(selected_flags, flags_for_stats)

    def extract_selected(
        self, record: SimpleFastxRecord, match: Union[ANPMatch, Match, None]
    ) -> Dict[str, FlagData]:
        assert match is not None
        flag_data: Dict[str, FlagData] = {}
        flag_data_all = self.extract_all(record, match)
        if self._selected_flags is not None:
            for flag, data in flag_data_all.items():
                if flag in self._selected_flags:
                    flag_data[flag] = data
        return flag_data

    def extract_all(
        self, record: SimpleFastxRecord, match: Union[ANPMatch, Match, None]
    ) -> Dict[str, FlagData]:
        if match is None:
            return {}
        flag_data: Dict[str, FlagData] = {}
        for gid in range(len(match.groups())):
            flag = self.__extract_single_flag(match, gid)
            flag_data.update([flag])
        return flag_data

    def __extract_single_flag(
        self,
        match: Union[ANPMatch, Match],
        gid: int,
        flag: Optional[Tuple[str, str]] = None,
    ) -> Tuple[str, FlagData]:
        if flag is None:
            flag = list(match.groupdict().items())[gid]
        return (flag[0], (flag[1], match.start(gid + 1), match.end(gid + 1)))

    def update(
        self, record: SimpleFastxRecord, flag_data: Dict[str, FlagData]
    ) -> SimpleFastxRecord:
        name, seq, _ = record
        name_bits = name.split(self._comment_space)
        for name, (flag, start, end) in flag_data.items():
            name_bits[0] += f"{self._flag_delim}{self._flag_delim}{name}"
            name_bits[0] += f"{self._flag_delim}{flag}"
        name = " ".join(name_bits)
        return (name, seq, None)


class FastqFlagExtractor(FastaFlagExtractor):

    extract_qual_flags: bool = True

    def __init__(
        self,
        selected_flags: Optional[List[str]] = None,
        flags_for_stats: Optional[List[str]] = None,
    ):
        super(FastqFlagExtractor, self).__init__(selected_flags, flags_for_stats)

    def extract_selected(
        self, record: SimpleFastxRecord, match: Union[ANPMatch, Match, None]
    ) -> Dict[str, FlagData]:
        assert match is not None
        name, seq, qual = record
        assert qual is not None
        flag_data = super(FastqFlagExtractor, self).extract_selected(record, match)
        if self.extract_qual_flags:
            flag_data = self.__add_qual_flags(flag_data, qual)
        return flag_data

    def extract_all(
        self, record: SimpleFastxRecord, match: Union[ANPMatch, Match, None]
    ) -> Dict[str, FlagData]:
        assert match is not None
        name, seq, qual = record
        assert qual is not None
        flag_data = super(FastqFlagExtractor, self).extract_all(record, match)
        if self.extract_qual_flags:
            flag_data = self.__add_qual_flags(flag_data, qual)
        return flag_data

    def __add_qual_flags(
        self, flag_data: Dict[str, FlagData], qual: str
    ) -> Dict[str, FlagData]:
        for name, (_, start, end) in list(flag_data.items()):
            flag = (f"{QFLAG_START}{name}", (qual[slice(start, end)], start, end))
            flag_data.update([flag])
        return flag_data

    def update(
        self, record: SimpleFastxRecord, flag_data: Dict[str, FlagData]
    ) -> SimpleFastxRecord:
        _, _, qual = record
        name, seq, _ = super(FastqFlagExtractor, self).update(record, flag_data)
        return (name, seq, qual)

    def apply_selection(self, flag_data: Dict[str, FlagData]) -> Dict[str, FlagData]:
        if self._selected_flags is None:
            return flag_data
        else:
            selected_flag_data = super(FastqFlagExtractor, self).apply_selection(
                flag_data
            )
            for name in self._selected_flags:
                name = f"{QFLAG_START}{name}"
                if name in flag_data.keys():
                    selected_flag_data[name] = flag_data[name]
            return selected_flag_data


def get_fastx_flag_extractor(fmt: FastxFormats) -> Type[ABCFlagExtractor]:
    """Retrieves appropriate flag extractor class."""
    if FastxFormats.FASTA == fmt:
        return FastaFlagExtractor
    elif FastxFormats.FASTQ == fmt:
        return FastqFlagExtractor
    else:
        return ABCFlagExtractor


class ABCFlagReader(ABCFlagBase):
    def __init__(self):
        super(ABCFlagReader, self).__init__()

    @abstractmethod
    def read(self, record: Any) -> Optional[Dict[str, FlagData]]:
        pass


class FastxFlagReader(ABCFlagReader):

    _flagstats: FlagStats

    def __init__(self, flags_for_stats: Optional[List[str]] = None):
        super(FastxFlagReader, self).__init__()
        self._flagstats = FlagStats(flags_for_stats)

    @property
    def flagstats(self):
        return self._flagstats

    def read(self, record: SimpleFastxRecord) -> Optional[Dict[str, FlagData]]:
        header = record[0]
        if self._comment_space in header:
            header = header.split(self._comment_space)[0]
        double_delim = f"{self._flag_delim}{self._flag_delim}"
        if double_delim not in header:
            return None
        flag_data: Dict[str, FlagData] = {}
        for flag in header.split(double_delim)[1:]:
            if self._flag_delim not in flag:
                continue
            name, value = flag.split(self._flag_delim)[:2]
            flag_data.update([(name, (value, -1, -1))])
        self._flagstats.update(flag_data)
        return flag_data


class FlagRegexes(object):

    _flag_regex: Dict[str, str]
    _flag_regex_compiled: Dict[str, Pattern]

    def __init__(self, pattern_list: List[str]):
        super(FlagRegexes, self).__init__()
        self._flag_regex = {}
        self.__init(pattern_list)
        self._flag_regex_compiled = {}
        self.__compile()

    def __init(self, pattern_list: List[str]) -> Dict[str, str]:
        self._flag_regex = {}
        for pattern in pattern_list:
            if "," not in pattern:
                continue
            exploded = pattern.split(",")
            self._flag_regex[exploded[0]] = ",".join(exploded[1:])
        return self._flag_regex

    def __compile(self) -> None:
        for name, regex in self._flag_regex.items():
            self._flag_regex_compiled[name] = re.compile(regex)

    def log(self) -> None:
        logging.info("[bold underline red]Flag regex[/]")
        for name, regex in self._flag_regex.items():
            logging.info(f"{name}-regex\t{regex}")

    def match(self, flags: Dict[str, FlagData]) -> bool:
        for name, regex in self._flag_regex_compiled.items():
            if name not in flags:
                return False
            match = re.match(regex, flags[name][0])
            if match is None:
                return False
        return True



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber import const, random, seqio
import shutil

TEST_ALPHABET = list("abcde")


def test_make_random_string():
    random_string = random.make_random_string(const.UT_RECORD_SEQ_LEN, TEST_ALPHABET)
    assert const.UT_RECORD_SEQ_LEN == len(random_string)
    assert all([c in TEST_ALPHABET for c in random_string])


def assert_fasta_record(
    record: seqio.SimpleFastaRecord, expected_name: str = const.UT_FLAG_NAME
) -> None:
    assert isinstance(record, tuple)
    assert isinstance(record[0], str)
    assert 3 == len(record)
    assert expected_name == record[0]
    assert isinstance(record[1], str)
    assert const.UT_RECORD_SEQ_LEN == len(record[1])
    assert all([c in random.DNA_ALPHABET for c in record[1]])
    assert record[2] is None


def test_make_fasta_record():
    assert_fasta_record(
        random.make_fasta_record(const.UT_FLAG_NAME, const.UT_RECORD_SEQ_LEN)
    )


def test_make_fasta_file():
    fastq = random.make_fasta_file(const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN)
    for i in range(len(fastq)):
        assert_fasta_record(fastq[i], f"record_{i+1}")


def assert_fastq_record(
    record: seqio.SimpleFastqRecord, expected_name: str = const.UT_FLAG_NAME
) -> None:
    assert isinstance(record, tuple)
    assert isinstance(record[0], str)
    assert 3 == len(record)
    assert expected_name == record[0]
    assert isinstance(record[1], str)
    assert const.UT_RECORD_SEQ_LEN == len(record[1])
    assert all([c in random.DNA_ALPHABET for c in record[1]])
    assert isinstance(record[2], str)
    assert const.UT_RECORD_SEQ_LEN == len(record[2])
    assert all([c in random.QUAL_ALPHABET for c in record[2]])


def test_make_fastq_record():
    assert_fastq_record(
        random.make_fastq_record(const.UT_FLAG_NAME, const.UT_RECORD_SEQ_LEN)
    )


def test_make_fastq_file():
    fastq = random.make_fastq_file(const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN)
    for i in range(len(fastq)):
        assert_fastq_record(fastq[i], f"record_{i+1}")


def assert_fastx_file(path: str, expected_fmt: const.FastxFormats) -> None:
    parser, fmt = seqio.get_fastx_parser(path)
    assert fmt is expected_fmt
    for record in parser:
        pass


def test_write_tmp_fasta_file():
    fpath, dpath = random.write_tmp_fasta_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    assert_fastx_file(fpath, const.FastxFormats.FASTA)
    zpath, _ = random.write_tmp_fasta_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN, dpath, gzipped=True
    )
    assert_fastx_file(zpath, const.FastxFormats.FASTA)
    shutil.rmtree(dpath)


def test_write_tmp_fastq_file():
    fpath, dpath = random.write_tmp_fastq_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    assert_fastx_file(fpath, const.FastxFormats.FASTQ)
    zpath, _ = random.write_tmp_fastq_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN, dpath, gzipped=True
    )
    assert_fastx_file(zpath, const.FastxFormats.FASTQ)
    shutil.rmtree(dpath)


def test_write_tmp_fastx_file():
    fpath, dpath = random.write_tmp_fasta_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    assert_fastx_file(fpath, const.FastxFormats.FASTA)
    zpath, _ = random.write_tmp_fasta_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN, dpath, gzipped=True
    )
    assert_fastx_file(zpath, const.FastxFormats.FASTA)
    fpath, _ = random.write_tmp_fastq_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN, dpath
    )
    assert_fastx_file(fpath, const.FastxFormats.FASTQ)
    zpath, _ = random.write_tmp_fastq_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN, dpath, gzipped=True
    )
    assert_fastx_file(zpath, const.FastxFormats.FASTQ)
    shutil.rmtree(dpath)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from abc import ABCMeta, abstractmethod
from fastx_barber.seqio import SimpleFastxRecord
from rich.progress import track  # type: ignore
from typing import Any, Dict, Iterator, List, Match, Optional, Pattern, Tuple, Union


class ANPMatch(object):
    __pattern: str
    __match: str
    __spans: List[Tuple[int, int]]
    __groups: List[str]
    __names: List[str]
    __groupdict: Dict[str, str]

    def __init__(
        self,
        match: str,
        span: Tuple[int, int],
        groups: List[Tuple[str, Tuple[int, int]]],
        pattern: str,
    ):
        super(ANPMatch, self).__init__()
        self.__pattern = pattern
        self.__match = match
        self.__spans = [span]
        self.__groups = [self.__match]
        self.__names = [""]
        self.__populate_groups(groups)

    def __populate_groups(self, groups: List[Tuple[str, Tuple[int, int]]]):
        for name, span in groups:
            self.__names.append(name)
            self.__spans.append(span)
            self.__groups.append(self.__match[span[0] : span[1]])

    @property
    def lastindex(self) -> int:
        return len(self.__groups)

    @property
    def lastgroup(self) -> str:
        return self.__names[-1]

    @property
    def pos(self) -> int:
        return self.start(None)

    @property
    def re(self):
        return self.__pattern

    def group(self, i: Optional[int]) -> str:
        if i is None:
            i = 0
        if i >= len(self.__spans):
            raise IndexError("no such group")
        return self.__groups[i]

    def groups(self) -> List[str]:
        return self.__groups[1:]

    def start(self, i: Optional[int]) -> int:
        if i is None:
            i = 0
        if i >= len(self.__spans):
            raise IndexError("no such group")
        return self.__spans[i][0]

    def end(self, i: Optional[int]) -> int:
        if i is None:
            i = 0
        if i >= len(self.__spans):
            raise IndexError("no such group")
        return self.__spans[i][1]

    def groupdict(self) -> Dict[str, str]:
        return dict(zip(self.__names[1:], self.__groups[1:]))


class AlphaNumericPattern(object):
    _pattern: str
    _length: int
    _groups: List[Tuple[str, Tuple[int, int]]]

    def __init__(self, pattern: str):
        super(AlphaNumericPattern, self).__init__()
        self._pattern = pattern
        self._groups = self.parse(pattern)
        self._length = sum([e - s for n, (s, e) in self._groups])

    @property
    def pattern(self):
        return self._pattern

    def match(self, query: str) -> Optional[ANPMatch]:
        if len(query) < self._length:
            return None
        return ANPMatch(
            query[: self._length], (0, self._length), self._groups, self._pattern
        )

    @staticmethod
    def remove_leading_digits(pattern: str) -> str:
        while pattern[0].isdigit():
            pattern = pattern[1:]
        return pattern

    @staticmethod
    def remove_trailing_alphas(pattern: str) -> str:
        while pattern[-1].isalpha():
            pattern = pattern[:-1]
        return pattern

    @staticmethod
    def parse(pattern: str) -> List[Tuple[str, Tuple[int, int]]]:
        assert pattern.isalnum()
        pattern = AlphaNumericPattern.remove_leading_digits(pattern)
        pattern = AlphaNumericPattern.remove_trailing_alphas(pattern)

        groups: Dict[str, Tuple[int, int]] = {}
        group_name = ""
        group_len = ""
        position = 0
        for c in pattern:
            if c.isalpha():
                if 0 != len(group_len):
                    assert group_name not in groups
                    groups[group_name] = (position, position + int(group_len))
                    position += int(group_len)
                    group_name = ""
                    group_len = ""
                group_name += c
            else:
                group_len += c
        groups[group_name] = (position, position + int(group_len))

        return sorted(groups.items(), key=lambda x: x[1][0])

    @staticmethod
    def to_regex(pattern: str) -> str:
        pattern_re = "^"
        for name, (start, end) in AlphaNumericPattern.parse(pattern):
            pattern_re += "(?<{name}>.{{end-start}})"
        return pattern_re

    def as_regex(self) -> str:
        return self.to_regex(self._pattern)


class ABCMatcher(metaclass=ABCMeta):
    """Record matcher abstract base class

    Extends:
        metaclass=ABCMeta

    Variables:
        _matched_count {int} -- number of matched records
        _unmatched_count {int} -- number of unmatched records
    """

    _matched_count: int = 0
    _unmatched_count: int = 0

    def __init__(self, pattern: Union[AlphaNumericPattern, Pattern]):
        """Init matcher class

        Arguments:
            pattern {Pattern} -- pattern to be matched
        """
        self._pattern = pattern

    @property
    def matched_count(self) -> int:
        return self._matched_count

    @property
    def unmatched_count(self) -> int:
        return self._unmatched_count

    @abstractmethod
    def do(self, record: Any) -> Tuple[Union[Optional[ANPMatch], Match], bool]:
        """Match a record with the provided pattern

        Decorators:
            abstractmethod

        Arguments:
            record {Any} -- record to be matched

        Returns:
            Tuple[Match, bool] -- match and matched status
        """
        pass


class FastxMatcher(ABCMatcher):
    def __init__(self, pattern: Union[AlphaNumericPattern, Pattern]):
        super(FastxMatcher, self).__init__(pattern)

    def do(
        self, record: SimpleFastxRecord
    ) -> Tuple[Union[Optional[ANPMatch], Match], bool]:
        name, seq, _ = record
        match = self._pattern.match(seq)
        matched = match is not None
        if matched:
            self._matched_count += 1
        else:
            self._unmatched_count += 1
        return (match, matched)


def search_needle(
    record: SimpleFastxRecord, needle: str, offset: int = 0
) -> Iterator[Tuple[int, int]]:
    header, seq, _ = record
    match_counter = offset
    for i in track(range(0, len(seq) - len(needle) + 1), description=header):
        if seq[i : (i + len(needle))] == needle:
            match_counter += 1
            yield (i, match_counter)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber.scripts import arguments
from fastx_barber.scripts import find_seq
from fastx_barber.scripts import flag, flag_extract
from fastx_barber.scripts import flag_filter, flag_regex, flag_split, flag_stats
from fastx_barber.scripts import match
from fastx_barber.scripts import trim, trim_len, trim_qual, trim_regex

__all__ = [
    "arguments",
    "find_seq",
    "flag",
    "flag_extract",
    "flag_filter",
    "flag_regex",
    "flag_split",
    "flag_stats",
    "match",
    "trim",
    "trim_len",
    "trim_qual",
    "trim_regex",
]



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from abc import ABCMeta, abstractmethod
from fastx_barber.const import FastxFormats
from fastx_barber.match import ANPMatch
from fastx_barber.qual import QualityIO
from fastx_barber.seqio import SimpleFastxRecord, SimpleFastqRecord
from typing import Any, List, Match, Tuple, Type, Union


class ABCTrimmer(metaclass=ABCMeta):
    """Record trimmer abstract base class

    Extends:
        metaclass=ABCMeta
    """

    @staticmethod
    @abstractmethod
    def trim_re(record: Any, match: Union[ANPMatch, Match, None]) -> Any:
        """Trim record using regexp match

        Decorators:
            staticmethod
            abstractmethod

        Arguments:
            record {Any} -- record to be trimmed
            match {Match} -- regexp match

        Returns:
            Any -- trimmed record
        """
        pass

    @staticmethod
    @abstractmethod
    def trim_len(record: Any, length: int, side: int) -> Any:
        """Trim record by length

        Decorators:
            staticmethod
            abstractmethod

        Arguments:
            record {Any} -- record to be trimmed
            length {int} -- length to be trimmed
            side {int} -- side to trim (5/3')

        Returns:
            Any -- trimmed record
        """
        pass


class FastaTrimmer(ABCTrimmer):
    def __init__(self):
        super(FastaTrimmer, self).__init__()

    @staticmethod
    def trim_re(
        record: SimpleFastxRecord, match: Union[ANPMatch, Match, None]
    ) -> SimpleFastxRecord:
        assert match is not None
        name, seq, _ = record
        seq = seq[: match.start(0)] + seq[match.end(0) :]
        return (name, seq, None)

    @staticmethod
    def trim_len(
        record: SimpleFastxRecord, length: int, side: int
    ) -> SimpleFastxRecord:
        if 5 == side:
            return (record[0], record[1][length:], None)
        elif 3 == side:
            return (record[0], record[1][:-length], None)
        else:
            raise Exception("Can trim only from 5' or 3' end.")


class FastqTrimmer(ABCTrimmer):
    def __init__(self):
        super(FastqTrimmer, self).__init__()

    @staticmethod
    def trim_re(
        record: SimpleFastxRecord, match: Union[ANPMatch, Match, None]
    ) -> SimpleFastxRecord:
        assert match is not None
        name, seq, qual = record
        assert qual is not None
        seq = seq[: match.start(0)] + seq[match.end(0) :]
        qual = qual[: match.start(0)] + qual[match.end(0) :]
        return (name, seq, qual)

    @staticmethod
    def trim_len(
        record: SimpleFastqRecord, length: int, side: int
    ) -> SimpleFastxRecord:
        if 5 == side:
            return (record[0], record[1][length:], record[2][length:])
        elif 3 == side:
            return (record[0], record[1][:-length], record[2][:-length])
        else:
            raise Exception("Can trim only from 5' or 3' end.")

    @staticmethod
    def __trim_qual_5(
        record: SimpleFastqRecord, qscore_thr: int, bases_qscores: List[int]
    ) -> Tuple[SimpleFastqRecord, int]:
        trimmed_length = 0
        while 0 < len(bases_qscores):
            if bases_qscores[0] < qscore_thr:
                record = (record[0], record[1][1:], record[2][1:])
                bases_qscores = bases_qscores[1:]
                trimmed_length += 1
            else:
                break
        return (record, trimmed_length)

    @staticmethod
    def __trim_qual_3(
        record: SimpleFastqRecord, qscore_thr: int, bases_qscores: List[int]
    ) -> Tuple[SimpleFastqRecord, int]:
        trimmed_length = 0
        while 0 < len(bases_qscores):
            if bases_qscores[-1] < qscore_thr:
                record = (record[0], record[1][:-1], record[2][:-1])
                bases_qscores = bases_qscores[:-1]
                trimmed_length += 1
            else:
                break
        return (record, trimmed_length)

    @staticmethod
    def trim_qual(
        record: SimpleFastqRecord, qscore_thr: int, side: int, qio: QualityIO
    ) -> Tuple[SimpleFastqRecord, int]:
        """Trim record by length

        Decorators:
            staticmethod
            abstractmethod

        Arguments:
            record {SimpleFastqRecord} -- record to be trimmed
            qscore_thr {int} -- qscore threshold, bases with lower qscore are trimmed
            side {int} -- side to trim (5/3')
            qio {QualityIO} -- QualityIO instance for qscore calculation

        Returns:
            SimpleFastqRecord -- trimmed record
        """
        bases_qscores = qio.phred_to_qscore(record[2])
        if 5 == side:
            return FastqTrimmer.__trim_qual_5(record, qscore_thr, bases_qscores)
        elif 3 == side:
            return FastqTrimmer.__trim_qual_3(record, qscore_thr, bases_qscores)
        else:
            raise Exception("Can trim only from 5' or 3' end.")


def get_fastx_trimmer(fmt: FastxFormats) -> Type[ABCTrimmer]:
    """Retrieves appropriate trimmer class."""
    if FastxFormats.FASTA == fmt:
        return FastaTrimmer
    elif FastxFormats.FASTQ == fmt:
        return FastqTrimmer
    else:
        return ABCTrimmer



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber.const import __version__
from fastx_barber.scripts import arguments as ap
from fastx_barber import scripts
import sys


def default_parser(*args) -> None:
    print("fbarber -h for usage details.")
    sys.exit()


def main():
    parser = argparse.ArgumentParser(
        description=f"""
Version:    {__version__}
Author:     Gabriele Girelli
Docs:       http://ggirelli.github.io/fastx-barber
Code:       http://github.com/ggirelli/fastx-barber

FASTX barber tools.
""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.set_defaults(parse=default_parser)
    parser = ap.add_version_option(parser)

    subparsers = parser.add_subparsers(
        title="sub-commands",
        help="Access the help page for a sub-command with: sub-command -h",
    )

    scripts.find_seq.init_parser(subparsers)
    scripts.flag.init_parser(subparsers)
    scripts.match.init_parser(subparsers)
    scripts.trim.init_parser(subparsers)

    args = parser.parse_args()
    args = args.parse(args)
    args.run(args)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber import qual
from fastx_barber.const import DEFAULT_PHRED_OFFSET

DEFAULT_FILTER_QUAL_FLAGS = ["flag,30,.2", "test,15,.1"]


def test_QualityIO():
    qio = qual.QualityIO()
    assert 32 == qio.phred_to_qscore("A")[0]
    assert 14 == qio.phred_to_qscore("/")[0]
    assert 60 == qio.phred_to_qscore("]")[0]
    assert 36 == qio.phred_to_qscore("E")[0]
    assert [32, 14, 60, 36] == qio.phred_to_qscore("A/]E")
    qio = qual.QualityIO(64)
    assert 32 - 31 == qio.phred_to_qscore("A")[0]
    assert 60 - 31 == qio.phred_to_qscore("]")[0]
    assert 36 - 31 == qio.phred_to_qscore("E")[0]
    assert [1, 29, 5] == qio.phred_to_qscore("A]E")


def test_QualityFilter():
    qf = qual.QualityFilter(30, 0.2)
    assert qf.qual_pass_filter("AAAA")
    assert not qf.qual_pass_filter("/AAA")
    assert qf.qual_pass_filter("AE]]]/")
    assert 3 == qf.parsed
    assert 2 == qf.passed


def validate_filters(filters):
    assert "qflag" in filters
    assert isinstance(filters["qflag"], qual.QualityFilter)
    assert 30 == filters["qflag"].min_qscore
    assert 0.2 == filters["qflag"].max_perc
    assert "qtest" in filters
    assert isinstance(filters["qtest"], qual.QualityFilter)
    assert 15 == filters["qtest"].min_qscore
    assert 0.1 == filters["qtest"].max_perc


def test_QualityFilter_init_flag_filters():
    filters = qual.QualityFilter.init_flag_filters(
        DEFAULT_FILTER_QUAL_FLAGS, DEFAULT_PHRED_OFFSET
    )
    validate_filters(filters)


def test_dummy_apply_filter_flag():
    assert qual.dummy_apply_filter_flag({}, {})


def test_apply_filter_flag():
    filters = qual.QualityFilter.init_flag_filters(
        DEFAULT_FILTER_QUAL_FLAGS, DEFAULT_PHRED_OFFSET
    )
    assert qual.apply_filter_flag(
        {"qflag": ("AAAA", -1, -1), "qtest": ("AAAAAAAAA", -1, -1)}, filters
    )
    assert not qual.apply_filter_flag(
        {"qflag": ("/AAA", -1, -1), "qtest": ("AAAAAAAAA", -1, -1)}, filters
    )
    assert not qual.apply_filter_flag(
        {"qflag": ("AAAA", -1, -1), "qtest": ("/AAAAAAAA", -1, -1)}, filters
    )
    assert not qual.apply_filter_flag(
        {"qflag": ("/AAA", -1, -1), "qtest": ("/AAAAAAAA", -1, -1)}, filters
    )
    assert qual.apply_filter_flag(
        {"qflag": ("/EEAAAA", -1, -1), "qtest": ("/EEAAAAAAAA", -1, -1)}, filters
    )


def test_setup_qual_filters():
    qff, ff = qual.setup_qual_filters(None, DEFAULT_PHRED_OFFSET)
    assert 0 == len(qff)
    assert qual.dummy_apply_filter_flag == ff
    qff, ff = qual.setup_qual_filters(DEFAULT_FILTER_QUAL_FLAGS, DEFAULT_PHRED_OFFSET)
    assert 2 == len(qff)
    assert qual.apply_filter_flag == ff
    validate_filters(qff)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber import io
from fastx_barber.const import BedExtension, BedRecord
import gzip
from typing import IO


class BedWriter(object):
    _OH: IO
    _n_fields: int
    __closed: bool

    def __init__(self, path: str, n_fields: int):
        super(BedWriter, self).__init__()
        base, ext, gzipped = io.is_gzipped(path)
        assert ext == BedExtension
        assert n_fields in [3, 4, 5, 6, 12]
        if gzipped:
            self._OH = gzip.open(path, "wt")
        else:
            self._OH = open(path, "w+")
        self.__closed = False
        self._n_fields = n_fields

    def __write_bed_record(self, record: BedRecord, n_fields: int):
        assert len(record) >= n_fields
        self._OH.write(f"{record[0]}")
        for i in range(1, n_fields):
            self._OH.write(f"\t{record[i]}")
        self._OH.write("\n")

    def do(self, record: BedRecord) -> None:
        assert not self.__closed
        self.__write_bed_record(record, self._n_fields)

    def close(self):
        self._OH.close()
        self.__closed = True



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import glob
import os
from rich.progress import track  # type: ignore
import tempfile
from typing import Optional, Tuple


DTEMP_PREFIX = "fbarber_tmp."


def check_tmp_dir(path: Optional[str] = None) -> str:
    if path is not None:
        assert os.path.isdir(path)
    else:
        path = tempfile.mkdtemp(prefix=DTEMP_PREFIX)
    return path


def splitext(path: str) -> Tuple[str, str]:
    base, ext = os.path.splitext(path)
    ext_final = ext
    while ".gz" == ext:
        base, ext = os.path.splitext(base)
        ext_final = ext + ext_final
    return (base, ext_final)


def is_gzipped(path: str) -> Tuple[str, str, bool]:
    """
    Returns:
        Tuple[str, str, bool] -- basename, extension, gzipped status
    """
    base, ext = splitext(path)
    gzipped = ext.endswith(".gz")
    if gzipped:
        ext = ".".join(ext.split(".")[:-1])
    return (base, ext, gzipped)


class ChunkMerger(object):
    _do_remove: bool = True
    _tempdir: Optional[tempfile.TemporaryDirectory]
    _split_by: Optional[str] = None

    def __init__(
        self,
        tempdir: Optional[tempfile.TemporaryDirectory] = None,
        split_by: Optional[str] = None,
    ):
        super(ChunkMerger, self).__init__()
        self._tempdir = tempdir
        self._split_by = split_by

    @property
    def do_remove(self) -> bool:
        return self._do_remove

    @do_remove.setter
    def do_remove(self, do_remove: bool) -> None:
        self._do_remove = do_remove

    def __merge_simple(
        self, path: str, last_chunk_id: int, desc: Optional[str] = None
    ) -> None:
        with open(path, "wb") as OH:
            for cid in track(
                range(1, last_chunk_id + 1), description=desc, transient=False
            ):
                chunk_path = f".tmp.chunk{cid}.{os.path.basename(path)}"
                if self._tempdir is not None:
                    chunk_path = os.path.join(self._tempdir.name, chunk_path)
                if not os.path.isfile(chunk_path):
                    continue
                with open(chunk_path, "rb") as CH:
                    OH.write(CH.read())
                if self._do_remove:
                    os.remove(chunk_path)

    def __merge_split(
        self, path: str, last_chunk_id: int, desc: Optional[str] = None
    ) -> None:
        output_dir = os.path.dirname(path)
        output_base = os.path.basename(path)
        for cid in track(
            range(1, last_chunk_id + 1), description=desc, transient=False
        ):
            chunk_path = f"{self._split_by}_split.*.tmp.chunk{cid}.{output_base}"
            if self._tempdir is not None:
                chunk_path = os.path.join(self._tempdir.name, chunk_path)
            flist = glob.glob(chunk_path)
            for fname in flist:
                split_value = (
                    os.path.basename(fname).split("_split.")[1].split(".tmp.chunk")[0]
                )
                with open(
                    os.path.join(
                        output_dir,
                        f"{self._split_by}_split.{split_value}.{output_base}",
                    ),
                    "ab",
                ) as OH:
                    with open(fname, "rb") as CH:
                        OH.write(CH.read())
                    if self._do_remove:
                        os.remove(fname)

    def do(self, path: str, last_chunk_id: int, desc: Optional[str] = None) -> None:
        if self._split_by is None:
            self.__merge_simple(path, last_chunk_id, desc)
        else:
            self.__merge_split(path, last_chunk_id, desc)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber import scriptio
from fastx_barber.const import PATTERN_EXAMPLE, FlagData
from fastx_barber.exception import enable_rich_assert
from fastx_barber.flag import (
    FastqFlagExtractor,
    FlagStats,
    get_fastx_flag_extractor,
)
from fastx_barber.io import ChunkMerger
from fastx_barber.match import AlphaNumericPattern, FastxMatcher
from fastx_barber.qual import setup_qual_filters
from fastx_barber.scriptio import get_handles, get_split_handles
from fastx_barber.scripts import arguments as ap
from fastx_barber.seqio import (
    get_fastx_format,
    SimpleFastxRecord,
    SimpleFastxWriter,
    SimpleSplitFastxWriter,
)
from fastx_barber.trim import get_fastx_trimmer
import joblib  # type: ignore
import logging
import regex as re  # type: ignore
from rich.logging import RichHandler  # type: ignore
import sys
from typing import Dict, List, Tuple, Union

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        "extract",
        description="Extract flags and trim the records of a FASTX file.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Extract flags and trim the records of a FASTX file.",
    )

    parser.add_argument(
        "input",
        type=str,
        metavar="in.fastx[.gz]",
        help="""Path to the fasta/q file to trim.""",
    )
    parser.add_argument(
        "output",
        type=str,
        metavar="out.fastx[.gz]",
        help="Path to fasta/q file where to write trimmed records. "
        + "Format will match the input.",
    )

    parser.add_argument(
        "--pattern",
        type=str,
        help="Pattern to match to reads and extract flagged groups. "
        + f"Remember to use quotes. Example: '{PATTERN_EXAMPLE}'",
    )

    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced = ap.add_unmatched_output_option(advanced)
    advanced = ap.add_flag_delim_option(advanced)
    advanced.add_argument(
        "--selected-flags",
        type=str,
        nargs="+",
        help="Space-separated names of flags to be extracted. "
        + "By default it extracts all flags.",
    )
    advanced = ap.add_flagstats_option(advanced)
    advanced = ap.add_split_by_option(advanced)
    advanced = ap.add_filter_qual_flags_option(advanced)
    advanced = ap.add_filter_qual_output_option(advanced)
    advanced = ap.add_phred_offset_option(advanced)
    advanced.add_argument(
        "--no-qual-flags",
        action="store_const",
        dest="qual_flags",
        const=False,
        default=True,
        help="Do not extract quality flags (when running on a fastq file).",
    )
    advanced.add_argument(
        "--simple-pattern",
        action="store_const",
        dest="simple_pattern",
        const=True,
        default=False,
        help="Parse pattern as 'simple' (alphanumeric) pattern.",
    )
    advanced = ap.add_comment_space_option(advanced)
    advanced = ap.add_compress_level_option(advanced)
    advanced = ap.add_log_file_option(advanced)

    advanced = ap.add_chunk_size_option(advanced)
    advanced = ap.add_threads_option(advanced)
    advanced = ap.add_tempdir_option(advanced)

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    assert 1 == len(args.flag_delim)
    args.threads = ap.check_threads(args.threads)
    args = scriptio.set_tempdir(args)

    if args.pattern is None:
        logging.info(
            "No pattern specified (--pattern), nothing to do. :person_shrugging:"
        )
        sys.exit()

    args.pattern = (
        AlphaNumericPattern(args.pattern)
        if args.simple_pattern
        else re.compile(args.pattern)
    )

    if args.log_file is not None:
        scriptio.add_log_file_handler(args.log_file)

    ap.log_args(args)
    logging.info("[bold underline red]Flag extraction[/]")
    if args.selected_flags is not None:
        logging.info(f"Selected flags\t{args.selected_flags}")
    logging.info(f"Flag stats\t{args.flagstats}")
    logging.info(f"Flag delim\t'{args.flag_delim}'")
    logging.info(f"Comment delim\t'{args.comment_space}'")
    logging.info(f"Quality flags\t{args.qual_flags}")
    if args.split_by is not None:
        logging.info(f"Split by\t'{args.split_by}'")

    return args


ChunkDetails = Tuple[int, int, int, FlagStats]


def run_chunk(
    chunk: List[SimpleFastxRecord],
    cid: int,
    args: argparse.Namespace,
) -> ChunkDetails:
    fmt, _ = get_fastx_format(args.input)
    OHC: Union[SimpleFastxWriter, SimpleSplitFastxWriter, None]
    FHC: Union[SimpleFastxWriter, SimpleSplitFastxWriter, None]
    OHC, UHC, FHC, filter_output_fun = (
        get_handles(fmt, cid, args)
        if args.split_by is None
        else get_split_handles(fmt, cid, args)
    )
    foutput = scriptio.get_output_fun(OHC, UHC)

    matcher = FastxMatcher(args.pattern)
    trimmer = get_fastx_trimmer(fmt)
    quality_flag_filters, filter_fun = setup_qual_filters(
        args.filter_qual_flags, args.phred_offset
    )

    flag_extractor = get_fastx_flag_extractor(fmt)(args.selected_flags, args.flagstats)
    flag_extractor.flag_delim = args.flag_delim
    flag_extractor.comment_space = args.comment_space
    if isinstance(flag_extractor, FastqFlagExtractor):
        flag_extractor.extract_qual_flags = args.qual_flags

    filtered_counter = 0
    for record in chunk:
        flags: Dict[str, FlagData] = {}
        match, matched = matcher.do(record)
        if matched:
            flags = flag_extractor.extract_all(record, match)
            flag_extractor.update_stats(flags)
            flags_selected = flag_extractor.apply_selection(flags)
            record = flag_extractor.update(record, flags_selected)
            record = trimmer.trim_re(record, match)
            pass_filters = filter_fun(flags, quality_flag_filters)
            if not pass_filters:
                filtered_counter += 1
                filter_output_fun(record, flags)
                continue
        foutput[matched](record, flags)

    SimpleFastxWriter.close_handle(OHC)
    SimpleFastxWriter.close_handle(UHC)
    SimpleFastxWriter.close_handle(FHC)

    return (
        filtered_counter,
        matcher.matched_count,
        len(chunk),
        flag_extractor.flagstats,
    )


def merge_chunk_details(chunk_details: List[ChunkDetails]) -> ChunkDetails:
    parsed_counter = 0
    matched_counter = 0
    filtered_counter = 0
    flagstats: FlagStats = FlagStats()
    for filtered, matched, parsed, stats in chunk_details:
        filtered_counter += filtered
        matched_counter += matched
        parsed_counter += parsed
        for flag_name, data in stats.items():
            for k, v in data.items():
                flagstats[flag_name][k] += v
    return (parsed_counter, matched_counter, filtered_counter, flagstats)


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    fmt, IH = scriptio.get_input_handler(args.input, args.chunk_size)

    quality_flag_filters, filter_fun = setup_qual_filters(
        args.filter_qual_flags, args.phred_offset, verbose=True
    )

    logging.info("[bold underline red]Running[/]")
    logging.info("Trimming and extracting flags...")
    chunk_details = joblib.Parallel(n_jobs=args.threads, verbose=10)(
        joblib.delayed(run_chunk)(chunk, cid, args) for chunk, cid in IH
    )
    logging.info("Merging subprocesses details...")
    n_parsed, n_matched, n_filtered, flagstats = merge_chunk_details(chunk_details)

    logging.info(
        f"{n_matched}/{n_parsed} ({n_matched/n_parsed*100:.2f}%) "
        + "records matched the pattern.",
    )
    if args.filter_qual_flags is not None and 0 != n_matched:
        logging.info(
            " ".join(
                (
                    f"{(n_matched-n_filtered)}/{n_matched}",
                    f"({(n_matched-n_filtered)/n_matched*100:.2f}%)",
                    "records passed the quality filters.",
                )
            )
        )

    if args.flagstats is not None:
        flagstats.export(args.output)

    logging.info("Merging batch output...")
    if args.unmatched_output is not None:
        merger = ChunkMerger(args.temp_dir, None)
        merger.do(args.unmatched_output, IH.last_chunk_id, "Writing unmatched records")
    merger = ChunkMerger(args.temp_dir, args.split_by)
    merger.do(args.output, IH.last_chunk_id, "Writing matched records")
    if args.filter_qual_output is not None:
        merger.do(args.filter_qual_output, IH.last_chunk_id, "Writing filtered records")

    logging.info("Done. :thumbs_up: :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber import match
import regex as re  # type: ignore


def test_AlphaNumericPattern():
    pass


def test_FastxMatcher():
    matcher = match.FastxMatcher(re.compile("GATC.{3}TTT"))
    assert matcher.do(("test", "GATCAAATTT", None))[1]
    assert not matcher.do(("test", "GATAAAATTT", None))[1]


def test_search_needle():
    record = ("test", "GATCACACATATATAGATCatcgatcagatcGATC", None)
    positions = [x for x in match.search_needle(record, "GATC", 0)]
    assert positions[0] == (0, 1)
    assert positions[1] == (15, 2)
    assert positions[2] == (31, 3)
    positions = [x for x in match.search_needle(record, "gatc", 0)]
    assert positions[0] == (22, 1)
    assert positions[1] == (27, 2)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber.const import FastxFormats
from fastx_barber.io import check_tmp_dir
from fastx_barber.seqio import (
    SimpleFastaRecord,
    SimpleFastaWriter,
    SimpleFastqRecord,
    SimpleFastqWriter,
)
from random import choice
import tempfile
from typing import List, Optional, Tuple

DNA_ALPHABET: List[str] = ["A", "T", "C", "G"]
QUAL_ALPHABET: List[str] = list("!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJ")


def make_random_string(length: int, alphabet: List[str] = DNA_ALPHABET) -> str:
    return "".join([choice(alphabet) for i in range(length)])


def make_fasta_record(name: str, length: int) -> SimpleFastaRecord:
    return (name, make_random_string(length), None)


def make_fasta_file(n_records: int, seq_length: int) -> List[SimpleFastaRecord]:
    records: List[SimpleFastaRecord] = []
    for i in range(n_records):
        records.append(make_fasta_record(f"record_{i+1}", seq_length))
    return records


def make_fastq_record(name: str, length: int) -> SimpleFastqRecord:
    return (
        *make_fasta_record(name, length)[:2],
        make_random_string(length, QUAL_ALPHABET),
    )


def make_fastq_file(n_records: int, seq_length: int) -> List[SimpleFastqRecord]:
    records: List[SimpleFastqRecord] = []
    for i in range(n_records):
        records.append(make_fastq_record(f"record_{i+1}", seq_length))
    return records


def mk_suffix(fmt: FastxFormats, gzipped):
    if fmt is FastxFormats.FASTA:
        suffix = ".fasta"
    else:
        suffix = ".fastq"
    if gzipped:
        suffix += ".gz"
    return suffix


def write_tmp_fasta_file(
    n_records: int,
    seq_length: int,
    tmp_dir: Optional[str] = None,
    gzipped: bool = False,
    compresslevel: int = 6,
) -> Tuple[str, str]:
    """Temporary FASTA file is not automatically deleted."""
    tmp_dir = check_tmp_dir(tmp_dir)
    generated_records = make_fasta_file(n_records, seq_length)
    _, tmp_fpath = tempfile.mkstemp(
        dir=tmp_dir, suffix=mk_suffix(FastxFormats.FASTA, gzipped), text=True
    )
    OH = SimpleFastaWriter(tmp_fpath)
    for record in generated_records:
        OH.write(record)
    return (tmp_fpath, tmp_dir)


def write_tmp_fastq_file(
    n_records: int,
    seq_length: int,
    tmp_dir: Optional[str] = None,
    gzipped: bool = False,
    compresslevel: int = 6,
) -> Tuple[str, str]:
    """Temporary FASTQ file is not automatically deleted."""
    tmp_dir = check_tmp_dir(tmp_dir)
    generated_records = make_fastq_file(n_records, seq_length)
    _, tmp_fpath = tempfile.mkstemp(
        dir=tmp_dir, suffix=mk_suffix(FastxFormats.FASTQ, gzipped), text=True
    )
    OH = SimpleFastqWriter(tmp_fpath)
    for record in generated_records:
        OH.write(record)
    return (tmp_fpath, tmp_dir)


def write_tmp_fastx_file(
    fmt: FastxFormats,
    n_records: int,
    seq_length: int,
    tmp_dir: Optional[str] = None,
    gzipped: bool = False,
    compresslevel: int = 6,
) -> Tuple[str, str]:
    """Temporary FASTX file is not automatically deleted."""
    if fmt is FastxFormats.FASTA:
        return write_tmp_fasta_file(
            n_records, seq_length, tmp_dir, gzipped, compresslevel
        )
    if fmt is FastxFormats.FASTQ:
        return write_tmp_fastq_file(
            n_records, seq_length, tmp_dir, gzipped, compresslevel
        )
    assert fmt is not FastxFormats.NONE



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber.const import DEFAULT_PHRED_OFFSET, FlagData, QFLAG_START
import logging
import numpy as np  # type: ignore
from typing import Callable, Dict, List, Tuple


class QualityIO(object):
    """docstring for QualityIO"""

    __phred_offset: int

    def __init__(self, phred_offset: int = 33):
        super(QualityIO, self).__init__()
        assert phred_offset in [33, 64]
        self.__phred_offset = phred_offset

    @property
    def phred_offset(self) -> int:
        return self.__phred_offset

    def phred_to_qscore(self, qual: str) -> List[int]:
        qscore = [ord(c) - self.__phred_offset for c in qual]
        assert not any(
            [q < 0 for q in qscore]
        ), f"phred offset of {self.__phred_offset} produces negative qscores"
        return qscore


class QualityFilter(QualityIO):
    """docstring for QualityFilter"""

    __min_qscore: int
    __max_perc: float

    _passed: int = 0
    _parsed: int = 0

    def __init__(
        self, min_qscore: int, max_perc: float, phred_offset: int = DEFAULT_PHRED_OFFSET
    ):
        super(QualityFilter, self).__init__(phred_offset)
        self.min_qscore = min_qscore
        self.max_perc = max_perc

    @property
    def min_qscore(self) -> int:
        return self.__min_qscore

    @min_qscore.setter
    def min_qscore(self, min_qscore: int) -> None:
        self.__min_qscore = min_qscore

    @property
    def max_perc(self) -> float:
        return self.__max_perc

    @max_perc.setter
    def max_perc(self, max_perc: float) -> None:
        assert max_perc >= 0 and max_perc <= 1
        self.__max_perc = max_perc

    @property
    def passed(self) -> int:
        return self._passed

    @property
    def parsed(self) -> int:
        return self._parsed

    def qual_pass_filter(self, qual: str) -> bool:
        """
        Arguments:
            qual {str} -- phred string

        Returns:
            bool -- whether the quality string passes the filter
        """
        has_passed = self.__qual_pass_filter(qual, self.__min_qscore, self.__max_perc)
        self._passed += has_passed
        self._parsed += 1
        return has_passed

    def __qual_pass_filter(self, qual: str, min_qscore: int, max_perc: float) -> bool:
        """
        Arguments:
            qual {str} -- phred string
            min_qscore {int} -- qscore threshold (lower qscores considered low quality)
            max_perc {float} -- max fraction (inclusive) of low quality bases
                                to pass the filter

        Returns:
            bool -- whether the quality string passes the filter
        """
        qscore = np.array(self.phred_to_qscore(qual))
        low_quality_fraction = (qscore < min_qscore).sum() / len(qual)
        return low_quality_fraction <= max_perc

    @staticmethod
    def init_flag_filters(
        filters: List[str], phred_offset: int
    ) -> Dict[str, "QualityFilter"]:
        filter_dict: Dict[str, QualityFilter] = {}
        for f in filters:
            flag, min_qscore_str, max_perc_str = f.split(",")
            min_qscore = int(min_qscore_str)
            max_perc = float(max_perc_str)
            filter_dict[f"{QFLAG_START}{flag}"] = QualityFilter(
                min_qscore, max_perc, phred_offset
            )
        return filter_dict


def dummy_apply_filter_flag(
    flag_data: Dict[str, FlagData], filters: Dict[str, QualityFilter]
) -> bool:
    return True


def apply_filter_flag(
    flag_data: Dict[str, FlagData], filters: Dict[str, QualityFilter]
) -> bool:
    """
    Arguments:
        flag_data {Dict[str, FlagData]} -- dict with flag name as key and data as value
        filters: {Dict[str, QualityFilter]} -- dict with flag name as key
                                               and filter as value

    Returns:
        bool -- whether the flags pass the filters
    """
    for flag, (qual, start, end) in flag_data.items():
        if not flag.startswith(QFLAG_START) or flag not in filters.keys():
            continue
        if not filters[flag].qual_pass_filter(qual):
            return False
    return True


def log_qual_filters(
    phred_offset: int, quality_flag_filters: Dict[str, QualityFilter]
) -> None:
    logging.info("[bold underline red]Quality filters[/]")
    logging.info(f"PHRED offset\t{phred_offset}")
    for name, f in quality_flag_filters.items():
        logging.info(f"{name}-filter\tmin_score={f.min_qscore} & max_perc={f.max_perc}")


def setup_qual_filters(
    filter_qual_flags: List[str], phred_offset: int, verbose: bool = False
) -> Tuple[Dict[str, QualityFilter], Callable]:
    quality_flag_filters: Dict[str, QualityFilter] = {}
    filter_fun = dummy_apply_filter_flag
    if filter_qual_flags is not None:
        quality_flag_filters = QualityFilter.init_flag_filters(
            filter_qual_flags, phred_offset
        )
        if verbose:
            log_qual_filters(phred_offset, quality_flag_filters)
        filter_fun = apply_filter_flag
    return (quality_flag_filters, filter_fun)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber import scriptio
from fastx_barber.exception import enable_rich_assert
from fastx_barber.const import FastxFormats
from fastx_barber.io import ChunkMerger
from fastx_barber.qual import QualityIO
from fastx_barber.scripts import arguments as ap
from fastx_barber.seqio import (
    FastxChunkedParser,
    get_fastx_format,
    get_fastx_parser,
    SimpleFastqRecord,
)
from fastx_barber.trim import FastqTrimmer
import joblib  # type: ignore
import logging
import numpy as np  # type: ignore
from rich.logging import RichHandler  # type: ignore
import sys
from typing import List, Tuple

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        "quality",
        description="Trim a FASTQ file by quality.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Trim a FASTQ file by quality.",
    )

    parser.add_argument(
        "input",
        type=str,
        metavar="in.fastq[.gz]",
        help="""Path to the fastq file to trim.""",
    )
    parser.add_argument(
        "output",
        type=str,
        metavar="out.fastq[.gz]",
        help="""Path to fastq file where to write
        trimmed records. Format will match the input.""",
    )

    parser.add_argument(
        "-q",
        "--qscore",
        type=int,
        default=0,
        help="QSCORE threshold. Any base with lower QSCORE is discarded. Default: 0",
    )
    parser.add_argument(
        "-s",
        "--side",
        type=int,
        default=5,
        choices=[3, 5],
        help="""Side to trim from. Either '5' for 5' (left/start),
        or '3' for 3' (right/end). Default: 5""",
    )

    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced = ap.add_phred_offset_option(advanced)
    advanced = ap.add_compress_level_option(advanced)
    advanced = ap.add_log_file_option(advanced)

    advanced = ap.add_chunk_size_option(advanced)
    advanced = ap.add_threads_option(advanced)
    advanced = ap.add_tempdir_option(advanced)

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    args.threads = ap.check_threads(args.threads)
    args = scriptio.set_tempdir(args)

    if 0 == args.qscore:
        logging.info(
            "Trimming QSCORE threshold (-q) equal to 0. "
            + "Nothing to do. :person_shrugging:"
        )
        sys.exit()

    if args.log_file is not None:
        scriptio.add_log_file_handler(args.log_file)

    return args


def run_chunk(
    chunk: List[SimpleFastqRecord],
    cid: int,
    args: argparse.Namespace,
) -> Tuple[int, int, List[int]]:
    fmt, _ = get_fastx_format(args.input)
    OHC = scriptio.get_chunk_handler(
        cid, fmt, args.output, args.compress_level, args.temp_dir
    )
    assert OHC is not None

    trimmer = FastqTrimmer()
    qio = QualityIO(args.phred_offset)

    trimmed_counter = 0
    untrimmed_counter = 0
    trimmed_length_list: List[int] = []
    for record in chunk:
        record, trimmed_length = trimmer.trim_qual(record, args.qscore, args.side, qio)
        trimmed_length_list.append(trimmed_length)
        if 0 < len(record[1]):
            OHC.write(record)
        trimmed_counter += trimmed_length != 0
        untrimmed_counter += trimmed_length == 0

    OHC.close()

    return (trimmed_counter, untrimmed_counter, trimmed_length_list)


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    logging.info("[bold underline red]General[/]")
    logging.info(f"Input\t\t{args.input}")
    logging.info(f"Trimming\tQSCORE < {args.qscore} from {args.side}'")
    logging.info(f"Threads\t\t{args.threads}")
    logging.info(f"Chunk size\t{args.chunk_size}")

    IH, fmt = get_fastx_parser(args.input)
    IH = FastxChunkedParser(IH, args.chunk_size)

    fmt, IH = scriptio.get_input_handler(args.input, args.chunk_size)
    assert FastxFormats.FASTQ == fmt, "Trimming by quality requires a fastq file."

    logging.info("[bold underline red]Running[/]")
    logging.info("Trimming...")
    output = joblib.Parallel(n_jobs=args.threads, verbose=10)(
        joblib.delayed(run_chunk)(
            chunk,
            cid,
            args,
        )
        for chunk, cid in IH
    )

    trimmed_length_list: List[int] = []
    trimmed_counter = 0
    parsed_counter = 0
    for trimmed, untrimmed, trimmed_lengths in output:
        trimmed_counter += trimmed
        parsed_counter += untrimmed
        trimmed_length_list.extend(trimmed_lengths)
    parsed_counter += trimmed_counter
    logging.info(
        " ".join(
            (
                f"{trimmed_counter}/{parsed_counter}",
                f"({trimmed_counter/parsed_counter*100:.2f}%) records trimmed.",
            )
        )
    )
    logging.info("Trimmed length statistics: min\tmean\tmedian\tmax")
    logging.info(
        "\t".join(
            (
                f"                           {min(trimmed_length_list)}",
                f"{np.mean(trimmed_length_list):.3f}",
                f"{np.median(trimmed_length_list):.3f}",
                f"{max(trimmed_length_list)}",
            )
        )
    )

    logging.info("Merging batch output...")
    merger = ChunkMerger(args.temp_dir, None)
    merger.do(args.output, IH.last_chunk_id, "Writing trimmed records")

    logging.info("Done. :thumbs_up: :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber.const import FastxFormats
from fastx_barber.seqio import (
    get_fastx_parser,
    get_fastx_writer,
    get_split_fastx_writer,
    FastxChunkedParser,
    SimpleFastxParser,
    SimpleFastxWriter,
    SimpleSplitFastxWriter,
)
import logging
import os
from rich.console import Console  # type: ignore
from rich.logging import RichHandler  # type: ignore
import tempfile
from typing import Callable, Dict, Optional, Tuple, Union


def set_tempdir(args: argparse.Namespace) -> argparse.Namespace:
    assert os.path.isdir(args.temp_dir), f"temporary folder not found: {args.temp_dir}"
    args.temp_dir = tempfile.TemporaryDirectory(
        prefix="fastx-barber.", dir=args.temp_dir
    )
    return args


def add_log_file_handler(path: str, logger_name: str = "") -> None:
    """Adds log file handler to logger.

    By defaults, adds the handler to the root logger.

    Arguments:
        path {str} -- path to output log file

    Keyword Arguments:
        logger_name {str} -- logger name (default: {""})
    """
    assert not os.path.isdir(path)
    log_dir = os.path.dirname(path)
    assert os.path.isdir(log_dir) or "" == log_dir
    fh = RichHandler(console=Console(file=open(path, mode="w+")), markup=True)
    fh.setLevel(logging.INFO)
    logging.getLogger(logger_name).addHandler(fh)
    logging.info(f"[green]Log to[/]\t\t{path}")


def get_input_handler(
    path: str, chunk_size: int
) -> Tuple[FastxFormats, SimpleFastxParser]:
    IH, fmt = get_fastx_parser(path)
    IH = FastxChunkedParser(IH, chunk_size)
    return (fmt, IH)


def get_chunk_tmp_path(
    cid: int,
    path: str,
    tempdir: Optional[tempfile.TemporaryDirectory] = None,
) -> str:
    chunk_path = f".tmp.chunk{cid}.{os.path.basename(path)}"
    if tempdir is not None:
        chunk_path = os.path.join(tempdir.name, os.path.basename(chunk_path))
    return chunk_path


def get_chunk_handler(
    cid: int,
    fmt: FastxFormats,
    path: Optional[str],
    compress_level: int,
    tempdir: Optional[tempfile.TemporaryDirectory] = None,
) -> Optional[SimpleFastxWriter]:
    if path is None:
        return None
    chunk_path = get_chunk_tmp_path(cid, path, tempdir)
    assert not os.path.isdir(path)
    return get_fastx_writer(fmt)(chunk_path, compress_level)


def get_split_chunk_handler(
    cid: int,
    fmt: FastxFormats,
    path: Optional[str],
    compress_level: int,
    split_by: str,
    tempdir: Optional[tempfile.TemporaryDirectory] = None,
) -> Optional[SimpleSplitFastxWriter]:
    if path is None:
        return None
    chunk_path = get_chunk_tmp_path(cid, path, tempdir)
    assert not os.path.isdir(path)
    return get_split_fastx_writer(fmt)(chunk_path, split_by, compress_level)


def get_output_fun(
    OH: Union[SimpleFastxWriter, SimpleSplitFastxWriter, None],
    UH: Optional[SimpleFastxWriter],
) -> Dict[bool, Callable]:
    """Prepare IO handlers.

    Arguments:
        OH {Optional[SimpleFastxWriter]} -- output buffer handler
        UH {Optional[SimpleFastxWriter]} -- unmatched output buffer handler
    """
    assert OH is not None
    if UH is not None:
        return {True: OH.write, False: UH.write}
    else:
        return {True: OH.write, False: lambda *x: None}


def get_qual_filter_handler(
    cid: int,
    fmt: FastxFormats,
    path: Optional[str],
    compress_level: int,
    tempdir: Optional[tempfile.TemporaryDirectory] = None,
) -> Tuple[Optional[SimpleFastxWriter], Callable]:
    FH = get_chunk_handler(cid, fmt, path, compress_level, tempdir)
    if FH is not None:
        assert fmt == FH.format, "format mismatch between input and requested output"
        return (FH, FH.write)
    else:
        return (FH, lambda *x: None)


def get_split_qual_filter_handler(
    cid: int,
    fmt: FastxFormats,
    path: Optional[str],
    compress_level: int,
    split_by: str,
    tempdir: Optional[tempfile.TemporaryDirectory] = None,
) -> Tuple[Optional[SimpleSplitFastxWriter], Callable]:
    FH = get_split_chunk_handler(cid, fmt, path, compress_level, split_by, tempdir)
    if FH is not None:
        assert fmt == FH.format, "format mismatch between input and requested output"
        return (FH, FH.write)
    else:
        return (FH, lambda *x: None)


def get_handles(
    fmt: FastxFormats, cid: int, args: argparse.Namespace
) -> Tuple[
    Optional[SimpleFastxWriter],
    Optional[SimpleFastxWriter],
    Optional[SimpleFastxWriter],
    Callable,
]:
    OHC = get_chunk_handler(cid, fmt, args.output, args.compress_level, args.temp_dir)
    assert OHC is not None
    UHC = get_chunk_handler(
        cid, fmt, args.unmatched_output, args.compress_level, args.temp_dir
    )
    FHC, filter_output_fun = get_qual_filter_handler(
        cid,
        fmt,
        args.filter_qual_output,
        args.compress_level,
        args.temp_dir,
    )
    return (OHC, UHC, FHC, filter_output_fun)


def get_split_handles(
    fmt: FastxFormats, cid: int, args: argparse.Namespace
) -> Tuple[
    Optional[SimpleSplitFastxWriter],
    Optional[SimpleFastxWriter],
    Optional[SimpleSplitFastxWriter],
    Callable,
]:
    OHC = get_split_chunk_handler(
        cid, fmt, args.output, args.compress_level, args.split_by, args.temp_dir
    )
    assert OHC is not None
    UHC = get_chunk_handler(
        cid, fmt, args.unmatched_output, args.compress_level, args.temp_dir
    )
    FHC, filter_output_fun = get_split_qual_filter_handler(
        cid,
        fmt,
        args.filter_qual_output,
        args.compress_level,
        args.split_by,
        args.temp_dir,
    )
    return (OHC, UHC, FHC, filter_output_fun)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber import scriptio
from fastx_barber.exception import enable_rich_assert
from fastx_barber.const import PATTERN_EXAMPLE
from fastx_barber.io import ChunkMerger
from fastx_barber.match import FastxMatcher, SimpleFastxRecord
from fastx_barber.scripts import arguments as ap
from fastx_barber.seqio import get_fastx_format
import joblib  # type: ignore
import logging
import regex  # type: ignore
from rich.logging import RichHandler  # type: ignore
import sys
from typing import List, Tuple

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        __name__.split(".")[-1],
        description="Scan a FASTX file for records matching a regular expression.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Scan a FASTX file for records matching a regular expression.",
    )

    parser.add_argument(
        "input",
        type=str,
        metavar="in.fastx[.gz]",
        help="""Path to the fasta/q file to scan for matches.""",
    )
    parser.add_argument(
        "output",
        type=str,
        metavar="out.fastx[.gz]",
        help="Path to fasta/q file where to write matching records. "
        + "Format will match the input.",
    )

    parser.add_argument(
        "--pattern",
        type=str,
        help="Pattern to match to reads. Remember to use quotes. "
        + f"Example: '{PATTERN_EXAMPLE}'",
    )

    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced = ap.add_unmatched_output_option(advanced)
    advanced = ap.add_compress_level_option(advanced)
    advanced = ap.add_log_file_option(advanced)

    advanced = ap.add_chunk_size_option(advanced)
    advanced = ap.add_threads_option(advanced)
    advanced = ap.add_tempdir_option(advanced)

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    args.threads = ap.check_threads(args.threads)
    args = scriptio.set_tempdir(args)

    if args.pattern is None:
        logging.info(
            "No pattern specified (--pattern), nothing to do. :person_shrugging:"
        )
        sys.exit()
    args.pattern = regex.compile(args.pattern)

    if args.log_file is not None:
        scriptio.add_log_file_handler(args.log_file)

    return args


def run_chunk(
    chunk: List[SimpleFastxRecord],
    cid: int,
    args: argparse.Namespace,
) -> Tuple[int, int]:
    fmt, _ = get_fastx_format(args.input)
    OHC = scriptio.get_chunk_handler(
        cid, fmt, args.output, args.compress_level, args.temp_dir
    )
    assert OHC is not None
    UHC = scriptio.get_chunk_handler(
        cid, fmt, args.unmatched_output, args.compress_level, args.temp_dir
    )
    foutput = scriptio.get_output_fun(OHC, UHC)

    matcher = FastxMatcher(args.pattern)

    for record in chunk:
        match, matched = matcher.do(record)
        foutput[matched](record)

    OHC.close()
    if UHC is not None:
        UHC.close()

    return (matcher.matched_count, len(chunk))


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    ap.log_args(args)

    fmt, IH = scriptio.get_input_handler(args.input, args.chunk_size)

    logging.info("[bold underline red]Running[/]")
    logging.info("Matching...")
    output = joblib.Parallel(n_jobs=args.threads, verbose=10)(
        joblib.delayed(run_chunk)(
            chunk,
            cid,
            args,
        )
        for chunk, cid in IH
    )

    parsed_counter = 0
    matched_counter = 0
    for matched, parsed in output:
        matched_counter += matched
        parsed_counter += parsed
    logging.info(
        " ".join(
            (
                f"{matched_counter}/{parsed_counter}",
                f"({matched_counter/parsed_counter*100:.2f}%)",
                "records matched the pattern.",
            )
        )
    )

    logging.info("Merging batch output...")
    merger = ChunkMerger(args.temp_dir, None)
    merger.do(args.output, IH.last_chunk_id, "Writing matched")
    if args.unmatched_output is not None:
        merger.do(args.unmatched_output, IH.last_chunk_id, "Writing unmatched")

    logging.info("Done. :thumbs_up: :smiley:")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastx_barber', 'fastx_barber.scripts', 'fastx_barber.tests']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.77,<2.0',
 'joblib>=0.16,<1.1',
 'numpy>=1.19.1,<2.0.0',
 'pandas>=1.1.2,<2.0.0',
 'pytest>=6.1.1,<7.0.0',
 'regex>=2020.7.14,<2022.0.0',
 'rich>=9.0,<10.0',
 'tqdm>=4.48.1,<5.0.0']

entry_points = \
{'console_scripts': ['fbarber = fastx_barber.scripts.barber:main']}

setup_kwargs = {
    'name': 'fastx-barber',
    'version': '0.1.3',
    'description': 'FASTX trimming tools',
    'long_description': "# fastx-barber\n\n[![DOI](https://zenodo.org/badge/281703558.svg)](https://zenodo.org/badge/latestdoi/281703558) ![](https://img.shields.io/librariesio/github/ggirelli/fastx-barber.svg?style=flat) ![](https://img.shields.io/github/license/ggirelli/fastx-barber.svg?style=flat)  \n![](https://github.com/ggirelli/fastx-barber/workflows/Python%20package/badge.svg?branch=main&event=push) ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/fastx-barber) ![PyPI - Format](https://img.shields.io/pypi/format/fastx-barber) ![PyPI - Status](https://img.shields.io/pypi/status/fastx-barber)  \n![](https://img.shields.io/github/release/ggirelli/fastx-barber.svg?style=flat) ![](https://img.shields.io/github/release-date/ggirelli/fastx-barber.svg?style=flat) ![](https://img.shields.io/github/languages/code-size/ggirelli/fastx-barber.svg?style=flat)  \n![](https://img.shields.io/github/watchers/ggirelli/fastx-barber.svg?label=Watch&style=social) ![](https://img.shields.io/github/stars/ggirelli/fastx-barber.svg?style=social)\n\n[PyPi](https://pypi.org/project/fastx-barber/) | [docs](https://ggirelli.github.io/fastx-barber/)\n\nA Python3.6.13+ package to trim and extract flags from FASTA  and FASTQ files.\n\n## Features (in short)\n\n* Works on both FASTA and FASTQ files.\n* Selects reads based on a pattern (regex).\n* Trims reads by pattern (regex), length, or single-base quality.\n* Extracts parts (flags) of reads based on a pattern and stores them in the read headers.\n* [Generates BED file with the locations of a substring](usage#find-sequence) in FASTX records.\n* Regular expression support [*fuzzy* matching](https://pypi.org/project/regex/#approximate-fuzzy-matching-hg-issue-12-hg-issue-41-hg-issue-109) (*fuzzy matching* might affect the barber's speed).\n* Parallelizes processing by splitting the fastx file in chunks.\n\nFor more available features, check out our [docs](https://ggirelli.github.io/fastx-barber/)!\n\n## Requirements\n\n`fastx-barber` has been tested with Python 3.6.13, 3.7, 3.8, and 3.9. We recommend installing it using `pipx` (see [below](https://github.com/ggirelli/fastx-barber#install)) to avoid dependency conflicts with other packages. The packages it depends on are listed in our [dependency graph](https://github.com/ggirelli/fastx-barber/network/dependencies). We use [`poetry`](https://github.com/python-poetry/poetry) to handle our dependencies.\n\n## Install\n\nWe recommend installing `fastx-barber` using [`pipx`](https://github.com/pipxproject/pipx). Check how to install `pipx` [here](https://github.com/pipxproject/pipx#install-pipx) if you don't have it yet!\n\nOnce you have `pipx` ready on your system, install the latest stable release of `fastx-barber` by running: `pipx install fastx-barber`. If you see the stars (✨ 🌟 ✨), then the installation went well!\n\n## Usage\n\nRun:\n\n* `fbarber` to access the barber's services.\n* `fbarber flag` to extract or manipulate read flags.\n* `fbarber match` to select reads based on a pattern (regular expression).\n* `fbarber trim` to trim your reads.\n\nAdd `-h` to see the full help page of a command!\n\n## Contributing\n\nWe welcome any contributions to `fastx-barber`. In short, we use [`black`](https://github.com/psf/black) to standardize code format. Any code change also needs to pass `mypy` checks. For more details, please refer to our [contribution guidelines](https://github.com/ggirelli/fastx-barber/blob/main/CONTRIBUTING.md) if this is your first time contributing! Also, check out our [code of conduct](https://github.com/ggirelli/fastx-barber/blob/main/CODE_OF_CONDUCT.md).\n\n## License\n\n`MIT License - Copyright (c) 2020 Gabriele Girelli`\n",
    'author': 'Gabriele Girelli',
    'author_email': 'gigi.ga90@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ggirelli/fastx-barber',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber import const, flag, match, random, seqio
import regex  # type: ignore
from typing import Dict


def test_FlagStats():
    fs = flag.FlagStats()
    fs.update({const.UT_FLAG_NAME: ("value", 0, 0)})
    assert 0 == fs.get_dataframe(const.UT_FLAG_NAME).shape[0]
    fs = flag.FlagStats(const.UT_FLAG_NAME)
    fs.update({const.UT_FLAG_NAME: ("value", 0, 0)})
    assert 1 == fs.get_dataframe(const.UT_FLAG_NAME).shape[0]


def assert_FastaFlagExtractor_update(
    fe: flag.FastaFlagExtractor,
    record: seqio.SimpleFastxRecord,
    flag_data: Dict[str, const.FlagData],
) -> None:
    expected_name = f"{record[0]}{fe.flag_delim}{fe.flag_delim}{const.UT_FLAG_NAME}"
    expected_name += f"{fe.flag_delim}{flag_data[const.UT_FLAG_NAME][0]}"
    updated_name, updated_seq, updated_qual = fe.update(record, flag_data)
    assert expected_name == updated_name
    assert record[1] == updated_seq
    assert updated_qual is None


def test_FastaFlagExtractor_noSelectedFlags_noStatFlags():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fasta_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastaFlagExtractor()
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        assert_FastaFlagExtractor_update(fe, record, flag_data)
        assert const.UT_FLAG_NAME not in fe.extract_selected(record, match_result)
    assert 0 == len(fe.flagstats.keys())


def test_FastaFlagExtractor_noStatFlags():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fasta_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastaFlagExtractor([const.UT_FLAG_NAME])
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        assert_FastaFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
        flag_data = fe.extract_selected(record, match_result)
        assert_FastaFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
    assert 0 == len(fe.flagstats.keys())


def test_FastaFlagExtractor():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fasta_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastaFlagExtractor([const.UT_FLAG_NAME], [const.UT_FLAG_NAME])
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        assert_FastaFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
        flag_data = fe.extract_selected(record, match_result)
        assert_FastaFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
    assert 1 == len(fe.flagstats.keys())


def test_FastaFlagExtractor_noSelectedFlags():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fasta_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastaFlagExtractor(None, [const.UT_FLAG_NAME])
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        assert_FastaFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
        assert const.UT_FLAG_NAME not in fe.extract_selected(record, match_result)
    assert 1 == len(fe.flagstats.keys())


def assert_FastqFlagExtractor_update(
    fe: flag.FastqFlagExtractor,
    record: seqio.SimpleFastxRecord,
    flag_data: Dict[str, const.FlagData],
) -> None:
    expected_name = f"{record[0]}{fe.flag_delim}{fe.flag_delim}{const.UT_FLAG_NAME}"
    expected_name += f"{fe.flag_delim}{flag_data[const.UT_FLAG_NAME][0]}"
    expected_name += f"{fe.flag_delim}{fe.flag_delim}q{const.UT_FLAG_NAME}"
    expected_name += f"{fe.flag_delim}{flag_data['q'+const.UT_FLAG_NAME][0]}"
    updated_name, updated_seq, updated_qual = fe.update(record, flag_data)
    assert expected_name == updated_name
    assert record[1] == updated_seq
    assert record[2] == updated_qual


def test_FastqFlagExtractor_noSelectedFlags_noStatFlags():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fastq_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastqFlagExtractor()
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        assert_FastqFlagExtractor_update(fe, record, flag_data)
        assert const.UT_FLAG_NAME not in fe.extract_selected(record, match_result)
    assert 0 == len(fe.flagstats.keys())


def test_FastqFlagExtractor_noStatFlags():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fastq_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastqFlagExtractor([const.UT_FLAG_NAME])
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        assert_FastqFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
        assert record[2][:8] == flag_data["q" + const.UT_FLAG_NAME][0]
        flag_data = fe.extract_selected(record, match_result)
        assert_FastqFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
        assert record[2][:8] == flag_data["q" + const.UT_FLAG_NAME][0]
    assert 0 == len(fe.flagstats.keys())


def test_FastqFlagExtractor():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fastq_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastqFlagExtractor([const.UT_FLAG_NAME], [const.UT_FLAG_NAME])
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        assert_FastqFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
        assert record[2][:8] == flag_data["q" + const.UT_FLAG_NAME][0]
        flag_data = fe.extract_selected(record, match_result)
        assert_FastqFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
        assert record[2][:8] == flag_data["q" + const.UT_FLAG_NAME][0]
    assert 1 == len(fe.flagstats.keys())


def test_FastqFlagExtractor_noSelectedFlags():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fastq_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastqFlagExtractor(None, [const.UT_FLAG_NAME])
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        assert_FastqFlagExtractor_update(fe, record, flag_data)
        assert record[1][:8] == flag_data[const.UT_FLAG_NAME][0]
        assert record[2][:8] == flag_data["q" + const.UT_FLAG_NAME][0]
        assert const.UT_FLAG_NAME not in fe.extract_selected(record, match_result)
    assert 1 == len(fe.flagstats.keys())


def test_get_fastx_flag_extractor():
    assert (
        flag.get_fastx_flag_extractor(const.FastxFormats.FASTA)
        is flag.FastaFlagExtractor
    )
    assert (
        flag.get_fastx_flag_extractor(const.FastxFormats.FASTQ)
        is flag.FastqFlagExtractor
    )
    assert (
        flag.get_fastx_flag_extractor(const.FastxFormats.NONE) is flag.ABCFlagExtractor
    )


def test_FastxFlagReader_fasta():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fasta_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastaFlagExtractor(None, [const.UT_FLAG_NAME])
    fr = flag.FastxFlagReader([const.UT_FLAG_NAME])
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        updated_record = fe.update(record, flag_data)
        read_flag_data = fr.read(updated_record)
        for k, v in flag_data.items():
            assert k in read_flag_data
            assert read_flag_data[k][0] == v[0]
        assert list(fe.flagstats.items()) == list(fr.flagstats.items())


def test_FastxFlagReader_fastq():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    generated_records = random.make_fastq_file(
        const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fe = flag.FastqFlagExtractor(None, [const.UT_FLAG_NAME])
    fr = flag.FastxFlagReader([const.UT_FLAG_NAME])
    for record in generated_records:
        match_result, matched = matcher.do(record)
        flag_data = fe.extract_all(record, match_result)
        fe.update_stats(flag_data)
        updated_record = fe.update(record, flag_data)
        read_flag_data = fr.read(updated_record)
        for k, v in flag_data.items():
            assert k in read_flag_data
            assert read_flag_data[k][0] == v[0]
        assert list(fe.flagstats.items()) == list(fr.flagstats.items())


def test_FlagRegexes_fasta():
    matcher = match.FastxMatcher(regex.compile(const.UT_FLAG_PATTERN))
    fe = flag.FastaFlagExtractor([const.UT_FLAG_NAME])
    record = ("fake", "ATCGATCGATCGATCGAT", None)
    match_result, matched = matcher.do(record)
    flag_data = fe.extract_all(record, match_result)
    fr = flag.FlagRegexes([f"{const.UT_FLAG_NAME},^AT.{{6}}$"])
    assert fr.match(flag_data)
    fr = flag.FlagRegexes([f"{const.UT_FLAG_NAME},^GT.{{6}}$"])
    assert not fr.match(flag_data)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber import scriptio
from fastx_barber.exception import enable_rich_assert
from fastx_barber.io import ChunkMerger
from fastx_barber.scripts import arguments as ap
from fastx_barber.seqio import SimpleFastxRecord, get_fastx_format
from fastx_barber.trim import get_fastx_trimmer
import joblib  # type: ignore
import logging
from rich.logging import RichHandler  # type: ignore
import sys
from typing import List, Tuple

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        "length",
        description="Trim a FASTX file by length.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Trim a FASTX file by length.",
    )

    parser.add_argument(
        "input",
        type=str,
        metavar="in.fastx[.gz]",
        help="""Path to the fasta/q file to trim.""",
    )
    parser.add_argument(
        "output",
        type=str,
        metavar="out.fastx[.gz]",
        help="""Path to fasta/q file where to write
        trimmed records. Format will match the input.""",
    )

    parser.add_argument(
        "-l",
        "--length",
        type=int,
        default=0,
        help="Number of bases to be trimmed. Default: 0",
    )
    parser.add_argument(
        "-s",
        "--side",
        type=int,
        default=5,
        choices=[3, 5],
        help="""Side to trim from. Either '5' for 5' (left/start),
        or '3' for 3' (right/end). Default: 5""",
    )

    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced = ap.add_compress_level_option(advanced)
    advanced = ap.add_log_file_option(advanced)

    advanced = ap.add_chunk_size_option(advanced)
    advanced = ap.add_threads_option(advanced)
    advanced = ap.add_tempdir_option(advanced)

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    args.threads = ap.check_threads(args.threads)
    args = scriptio.set_tempdir(args)

    if 0 == args.length:
        logging.info(
            "Trimming length (--length) equal to 0. Nothing to do. :person_shrugging:"
        )
        sys.exit()

    if args.log_file is not None:
        scriptio.add_log_file_handler(args.log_file)

    return args


def run_chunk(
    chunk: List[SimpleFastxRecord],
    cid: int,
    args: argparse.Namespace,
) -> Tuple[int, int]:
    fmt, _ = get_fastx_format(args.input)
    OHC = scriptio.get_chunk_handler(
        cid, fmt, args.output, args.compress_level, args.temp_dir
    )
    assert OHC is not None

    trimmer = get_fastx_trimmer(fmt)

    skipped_short_counter = 0
    trimmed_counter = 0
    for record in chunk:
        if len(record[1]) <= args.length:
            skipped_short_counter += 1
            continue
        record = trimmer.trim_len(record, args.length, args.side)
        OHC.write(record)
        trimmed_counter += 1

    OHC.close()

    return (trimmed_counter, skipped_short_counter)


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    logging.info("[bold underline red]General[/]")
    logging.info(f"Input\t\t{args.input}")
    logging.info(f"Trimming\t{args.length} nt from {args.side}'")
    logging.info(f"Threads\t\t{args.threads}")
    logging.info(f"Chunk size\t{args.chunk_size}")

    fmt, IH = scriptio.get_input_handler(args.input, args.chunk_size)

    logging.info("[bold underline red]Running[/]")
    logging.info("Trimming...")
    output = joblib.Parallel(n_jobs=args.threads, verbose=10)(
        joblib.delayed(run_chunk)(
            chunk,
            cid,
            args,
        )
        for chunk, cid in IH
    )

    parsed_counter = 0
    trimmed_counter = 0
    for trimmed, skipped_short in output:
        trimmed_counter += trimmed
        parsed_counter += skipped_short
    parsed_counter += trimmed_counter
    logging.info(
        " ".join(
            (
                f"{trimmed_counter}/{parsed_counter}",
                f"({trimmed_counter/parsed_counter*100:.2f}%)",
                "records trimmed.",
            )
        )
    )

    logging.info("Merging batch output...")
    merger = ChunkMerger(args.temp_dir, None)
    merger.do(args.output, IH.last_chunk_id, "Writing trimmed records")

    logging.info("Done. :thumbs_up: :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber import const, io, random, seqio
import os
import shutil
import tempfile


def test_get_fastx_format():
    assert (const.FastxFormats.FASTA, True) == seqio.get_fastx_format("fake.fasta.gz")
    assert (const.FastxFormats.FASTA, False) == seqio.get_fastx_format("fake.fasta")
    assert (const.FastxFormats.FASTQ, True) == seqio.get_fastx_format("fake.fastq.gz")
    assert (const.FastxFormats.FASTQ, False) == seqio.get_fastx_format("fake.fastq")
    assert (const.FastxFormats.FASTA, True) == seqio.get_fastx_format("fake.fa.gz")
    assert (const.FastxFormats.FASTA, False) == seqio.get_fastx_format("fake.fa")
    assert (const.FastxFormats.FASTQ, True) == seqio.get_fastx_format("fake.fq.gz")
    assert (const.FastxFormats.FASTQ, False) == seqio.get_fastx_format("fake.fq")
    try:
        seqio.get_fastx_format("fake.txt.gz")
    except AssertionError:
        pass
    try:
        seqio.get_fastx_format("fake.txt")
    except AssertionError:
        pass


def test_get_fastx_parser():
    fapath, dpath = random.write_tmp_fastx_file(
        const.FastxFormats.FASTA, const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    for record in seqio.get_fastx_parser(fapath)[0]:
        pass
    fqpath, _ = random.write_tmp_fastx_file(
        const.FastxFormats.FASTA, const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN, dpath
    )
    for record in seqio.get_fastx_parser(fqpath)[0]:
        pass
    zfapath, _ = random.write_tmp_fastx_file(
        const.FastxFormats.FASTA,
        const.UT_N_RECORDS,
        const.UT_RECORD_SEQ_LEN,
        dpath,
        gzipped=True,
    )
    for record in seqio.get_fastx_parser(zfapath)[0]:
        pass
    zfqpath, _ = random.write_tmp_fastx_file(
        const.FastxFormats.FASTA,
        const.UT_N_RECORDS,
        const.UT_RECORD_SEQ_LEN,
        dpath,
        gzipped=True,
    )
    for record in seqio.get_fastx_parser(zfqpath)[0]:
        pass
    shutil.rmtree(dpath)


def test_FastxChunkedParser():
    fapath, dpath = random.write_tmp_fastx_file(
        const.FastxFormats.FASTA, const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN
    )
    fachunks = [
        c
        for c in seqio.FastxChunkedParser(
            seqio.get_fastx_parser(fapath)[0], const.UT_CHUNK_SIZE
        )
    ]
    assert 13 == len(fachunks)
    fqpath, _ = random.write_tmp_fastx_file(
        const.FastxFormats.FASTA, const.UT_N_RECORDS, const.UT_RECORD_SEQ_LEN, dpath
    )
    fqchunks = [
        c
        for c in seqio.FastxChunkedParser(
            seqio.get_fastx_parser(fqpath)[0], const.UT_CHUNK_SIZE
        )
    ]
    assert 13 == len(fqchunks)
    zfapath, _ = random.write_tmp_fastx_file(
        const.FastxFormats.FASTA,
        const.UT_N_RECORDS,
        const.UT_RECORD_SEQ_LEN,
        dpath,
        gzipped=True,
    )
    fachunks = [
        c
        for c in seqio.FastxChunkedParser(
            seqio.get_fastx_parser(zfapath)[0], const.UT_CHUNK_SIZE
        )
    ]
    assert 13 == len(fachunks)
    zfqpath, _ = random.write_tmp_fastx_file(
        const.FastxFormats.FASTA,
        const.UT_N_RECORDS,
        const.UT_RECORD_SEQ_LEN,
        dpath,
        gzipped=True,
    )
    fqchunks = [
        c
        for c in seqio.FastxChunkedParser(
            seqio.get_fastx_parser(zfqpath)[0], const.UT_CHUNK_SIZE
        )
    ]
    assert 13 == len(fqchunks)
    shutil.rmtree(dpath)


def test_SimpleFastaWriter():
    tmp_dir = io.check_tmp_dir()
    generated_records = random.make_fasta_file(const.UT_RECORD_SEQ_LEN, 200)
    _, fpath = tempfile.mkstemp(
        dir=tmp_dir, suffix=random.mk_suffix(const.FastxFormats.FASTA, False), text=True
    )
    OH = seqio.SimpleFastaWriter(fpath)
    for record in generated_records:
        OH.write(record)
    OH.close()
    written_records = [r for r in seqio.get_fastx_parser(fpath)[0]]
    for i in range(len(generated_records)):
        assert generated_records[i] == written_records[i]
    shutil.rmtree(tmp_dir)


def test_SimpleFastqWriter():
    tmp_dir = io.check_tmp_dir()
    generated_records = random.make_fastq_file(const.UT_RECORD_SEQ_LEN, 200)
    _, fpath = tempfile.mkstemp(
        dir=tmp_dir, suffix=random.mk_suffix(const.FastxFormats.FASTQ, False), text=True
    )
    OH = seqio.SimpleFastqWriter(fpath)
    for record in generated_records:
        OH.write(record)
    OH.close()
    written_records = [r for r in seqio.get_fastx_parser(fpath)[0]]
    for i in range(len(generated_records)):
        assert generated_records[i] == written_records[i]
    shutil.rmtree(tmp_dir)


def test_get_fastx_writer():
    assert seqio.get_fastx_writer(const.FastxFormats.FASTA) is seqio.SimpleFastaWriter
    assert seqio.get_fastx_writer(const.FastxFormats.FASTQ) is seqio.SimpleFastqWriter
    assert seqio.get_fastx_writer(const.FastxFormats.NONE) is seqio.SimpleFastxWriter


def test_SimpleSplitFastaWriter():
    tmp_dir = io.check_tmp_dir()
    generated_records = random.make_fasta_file(const.UT_RECORD_SEQ_LEN, 200)
    _, fpath = tempfile.mkstemp(
        dir=tmp_dir, suffix=random.mk_suffix(const.FastxFormats.FASTA, False), text=True
    )
    OH = seqio.SimpleSplitFastaWriter(fpath, "first")
    for record in generated_records:
        OH.write(record, {"first": (record[1][0], 0, 0)})
    for c in random.DNA_ALPHABET:
        parser, _ = seqio.get_fastx_parser(
            os.path.join(tmp_dir, f"first_split.{c}.{os.path.basename(fpath)}")
        )
        for record in parser:
            assert c == record[1][0]
    shutil.rmtree(tmp_dir)


def test_SimpleSplitFastqWriter():
    tmp_dir = io.check_tmp_dir()
    generated_records = random.make_fastq_file(const.UT_RECORD_SEQ_LEN, 200)
    _, fpath = tempfile.mkstemp(
        dir=tmp_dir, suffix=random.mk_suffix(const.FastxFormats.FASTQ, False), text=True
    )
    OH = seqio.SimpleSplitFastqWriter(fpath, "first")
    for record in generated_records:
        OH.write(record, {"first": (record[1][0], 0, 0)})
    for c in random.DNA_ALPHABET:
        parser, _ = seqio.get_fastx_parser(
            os.path.join(tmp_dir, f"first_split.{c}.{os.path.basename(fpath)}")
        )
        for record in parser:
            assert c == record[1][0]
    shutil.rmtree(tmp_dir)


def test_get_split_fastx_writer():
    assert (
        seqio.get_split_fastx_writer(const.FastxFormats.FASTA)
        is seqio.SimpleSplitFastaWriter
    )
    assert (
        seqio.get_split_fastx_writer(const.FastxFormats.FASTQ)
        is seqio.SimpleSplitFastqWriter
    )
    assert (
        seqio.get_split_fastx_writer(const.FastxFormats.NONE)
        is seqio.SimpleSplitFastxWriter
    )



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber import bedio
import os

bed_data = [
    ("chr1", 1, 10, "row1", 3.0, "+", None, None, None, None, None, None),
    ("chr1", 12, 102, "row2", 5.0, "-", None, None, None, None, None, None),
    ("chr4", 10, 12, "row3", 3.7, "+", None, None, None, None, None, None),
]


def test_BedWriter_3fields():
    bed_path = "test.bed"
    bw = bedio.BedWriter(bed_path, 3)
    for record in bed_data:
        bw.do(record)
    bw.close()

    with open(bed_path) as IH:
        bed_content = IH.readlines()
        assert len(bed_data) == len(bed_content)
        for i in range(len(bed_data)):
            record = bed_content[i].strip().split()
            assert 3 == len(record)
            assert bed_data[i][0] == record[0]
            assert bed_data[i][1] == int(record[1])
            assert bed_data[i][2] == int(record[2])

    os.remove(bed_path)


def test_BedWriter_4fields():
    bed_path = "test.bed"
    bw = bedio.BedWriter(bed_path, 4)
    for record in bed_data:
        bw.do(record)
    bw.close()

    with open(bed_path) as IH:
        bed_content = IH.readlines()
        assert len(bed_data) == len(bed_content)
        for i in range(len(bed_data)):
            record = bed_content[i].strip().split()
            assert 4 == len(record)
            assert bed_data[i][0] == record[0]
            assert bed_data[i][1] == int(record[1])
            assert bed_data[i][2] == int(record[2])
            assert bed_data[i][3] == record[3]

    os.remove(bed_path)


def test_BedWriter_5fields():
    bed_path = "test.bed"
    bw = bedio.BedWriter(bed_path, 5)
    for record in bed_data:
        bw.do(record)
    bw.close()

    with open(bed_path) as IH:
        bed_content = IH.readlines()
        assert len(bed_data) == len(bed_content)
        for i in range(len(bed_data)):
            record = bed_content[i].strip().split()
            assert 5 == len(record)
            assert bed_data[i][0] == record[0]
            assert bed_data[i][1] == int(record[1])
            assert bed_data[i][2] == int(record[2])
            assert bed_data[i][3] == record[3]
            assert bed_data[i][4] == float(record[4])

    os.remove(bed_path)


def test_BedWriter_6fields():
    bed_path = "test.bed"
    bw = bedio.BedWriter(bed_path, 6)
    for record in bed_data:
        bw.do(record)
    bw.close()

    with open(bed_path) as IH:
        bed_content = IH.readlines()
        assert len(bed_data) == len(bed_content)
        for i in range(len(bed_data)):
            record = bed_content[i].strip().split()
            assert 6 == len(record)
            assert bed_data[i][0] == record[0]
            assert bed_data[i][1] == int(record[1])
            assert bed_data[i][2] == int(record[2])
            assert bed_data[i][3] == record[3]
            assert bed_data[i][4] == float(record[4])
            assert bed_data[i][5] == record[5]

    os.remove(bed_path)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from fastx_barber.const import __version__
from fastx_barber import const, scripts
from fastx_barber import bedio, io, scriptio, seqio
from fastx_barber import flag, match, qual, trim

__all__ = [
    "__version__",
    "const",
    "scripts",
    "bedio",
    "io",
    "scriptio",
    "seqio",
    "flag",
    "match",
    "qual",
    "trim",
]



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
from fastx_barber import scriptio
from fastx_barber.exception import enable_rich_assert
from fastx_barber.flag import FastxFlagReader, FlagStats
from fastx_barber.match import SimpleFastxRecord
from fastx_barber.scripts import arguments as ap
import joblib  # type: ignore
import logging
from rich.logging import RichHandler  # type: ignore
import sys
from typing import List

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[RichHandler(markup=True, rich_tracebacks=True)],
)


def init_parser(subparsers: argparse._SubParsersAction) -> argparse.ArgumentParser:
    parser = subparsers.add_parser(
        "stats",
        description="Calculate FASTX flag stats.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        help="Calculate FASTX flag stats.",
    )

    parser.add_argument(
        "input",
        type=str,
        metavar="in.fastx[.gz]",
        help="Path to the fasta/q file to scan for matches.",
    )

    parser = ap.add_version_option(parser)

    advanced = parser.add_argument_group("advanced arguments")
    advanced = ap.add_flag_delim_option(advanced)
    advanced = ap.add_comment_space_option(advanced)
    advanced = ap.add_flagstats_option(advanced)

    advanced = ap.add_compress_level_option(advanced)
    advanced = ap.add_log_file_option(advanced)
    advanced = ap.add_chunk_size_option(advanced)
    advanced = ap.add_threads_option(advanced)
    advanced = ap.add_tempdir_option(advanced)

    parser.set_defaults(parse=parse_arguments, run=run)

    return parser


@enable_rich_assert
def parse_arguments(args: argparse.Namespace) -> argparse.Namespace:
    args.threads = ap.check_threads(args.threads)
    args = scriptio.set_tempdir(args)

    if args.flagstats is None:
        logging.info(
            "No flag specified (--flagstats), nothing to do. :person_shrugging:"
        )
        sys.exit()

    if args.log_file is not None:
        scriptio.add_log_file_handler(args.log_file)

    return args


def run_chunk(
    chunk: List[SimpleFastxRecord],
    cid: int,
    args: argparse.Namespace,
) -> None:
    fmt, IH = scriptio.get_input_handler(args.input, args.chunk_size)

    flag_reader = FastxFlagReader(args.flagstats)
    flag_reader.flag_delim = args.flag_delim
    flag_reader.comment_space = args.comment_space

    for record in chunk:
        flag_reader.read(record)

    return flag_reader.flagstats


def merge_flagstats(chunk_details: List[FlagStats]) -> FlagStats:
    flagstats: FlagStats = FlagStats()
    for stats in chunk_details:
        for flag_name, data in stats.items():
            for k, v in data.items():
                flagstats[flag_name][k] += v
    return flagstats


@enable_rich_assert
def run(args: argparse.Namespace) -> None:
    logging.info("[bold underline red]General[/]")
    logging.info(f"Input\t\t{args.input}")
    logging.info(f"Threads\t\t{args.threads}")
    logging.info(f"Chunk size\t{args.chunk_size}")
    logging.info("[bold underline red]Flag extraction[/]")
    logging.info(f"Flag delim\t'{args.flag_delim}'")
    logging.info(f"Comment delim\t'{args.comment_space}'")
    logging.info(f"Flag stats\t'{args.flagstats}'")

    fmt, IH = scriptio.get_input_handler(args.input, args.chunk_size)

    logging.info("[bold underline red]Running[/]")
    logging.info("Matching...")
    chunk_details = joblib.Parallel(n_jobs=args.threads, verbose=10)(
        joblib.delayed(run_chunk)(
            chunk,
            cid,
            args,
        )
        for chunk, cid in IH
    )

    merge_flagstats(chunk_details).export(args.input)

    logging.info("Done. :thumbs_up: :smiley:")
